from uuid import uuid4
from django.contrib.auth import get_user_model
from django.urls import reverse

from rest_framework.test import APIClient
from rest_framework import status

from jaseci.utils.utils import TestCaseHelper
from django.test import TestCase


# Consts for url
CREATE_USER_URL = reverse("user_api:create")
TOKEN_URL = reverse("user_api:token")
MANAGE_URL = reverse("user_api:manage")
LOGOUT_EVERYONE_URL = reverse("user_api:logout_everyone")
PASSWORD_RESET_URL = reverse("user_api:password_reset:reset-password-request")
# Alias for create user
create_user = get_user_model().objects.create_user
create_superuser = get_user_model().objects.create_superuser
get_user = get_user_model().objects.get


class UserApiPublicTests(TestCaseHelper, TestCase):
    """Tests for user API (public)"""

    def setUp(self):
        super().setUp()
        self.client = APIClient()

    def tearDown(self):
        super().tearDown()

    def test_create_valid_user_success(self):
        """Test creating user with valid payload is successful"""
        payload = {
            "email": "jscitest_test2@jaseci.com",
            "password": "testpass",
            "name": "name",
        }
        res = self.client.post(CREATE_USER_URL, payload)

        self.assertEqual(res.status_code, status.HTTP_201_CREATED)
        user = get_user_model().objects.get(**res.data)
        self.assertTrue(user.check_password(payload["password"]))
        self.assertNotIn("password", res.data)
        user.delete()

    def test_first_user_is_super(self):
        """Test first user created its superuser"""
        payload = {
            "email": "jscitest_test2@jaseci.com",
            "password": "testpass",
            "name": "name",
        }
        res = self.client.post(CREATE_USER_URL, payload)

        self.assertEqual(res.status_code, status.HTTP_201_CREATED)
        user = get_user_model().objects.get(**res.data)
        self.assertTrue(user.is_admin)
        payload = {
            "email": "jscitest_test4@jaseci.com",
            "password": "testpass",
            "name": "name",
        }
        res = self.client.post(CREATE_USER_URL, payload)

        self.assertEqual(res.status_code, status.HTTP_201_CREATED)
        user2 = get_user_model().objects.get(**res.data)
        self.assertFalse(user2.is_admin)
        user.delete()
        user2.delete()

    def test_create_activated_user_success(self):
        """Test creating user with valid payload is successful"""
        payload = {
            "email": "jscitest_test2@jaseci.com",
            "password": "testpass",
            "name": "name",
            "is_activated": True,
        }
        res = self.client.post(CREATE_USER_URL, payload)

        self.assertEqual(res.status_code, status.HTTP_201_CREATED)
        user = get_user_model().objects.get(**res.data)
        self.assertTrue(user.check_password(payload["password"]))
        self.assertTrue(user.is_activated)
        self.assertNotIn("password", res.data)
        user.delete()

    def test_user_exists(self):
        """Test creating a user that already exists fails"""
        payload = {"email": "jscitest_test2@jaseci.com", "password": "testpass"}
        create_user(**payload)
        res = self.client.post(CREATE_USER_URL, payload)

        self.assertEqual(res.status_code, status.HTTP_400_BAD_REQUEST)
        get_user(email=payload["email"]).delete()

    def test_password_too_short(self):
        """Test that passwords less than 8 characters fail"""
        payload = {"email": "jscitest_test2@jaseci.com", "password": "pw"}
        res = self.client.post(CREATE_USER_URL, payload)

        self.assertEqual(res.status_code, status.HTTP_400_BAD_REQUEST)
        user_exists = get_user_model().objects.filter(email=payload["email"]).exists()
        self.assertFalse(user_exists)

    def test_logout_everyone(self):
        """Test admin level logout everyone"""
        payload = {"email": "jscitest_test@jaseci.com", "password": "testpass"}
        user = create_user(**payload)
        user.is_activated = True
        user.save()
        payload2 = {"email": "super@jaseci.com", "password": "testpass"}
        superuser = create_superuser(**payload2)
        superuser.save()

        res = self.client.post(TOKEN_URL, payload)
        token = res.data["token"]
        res = self.client.post(TOKEN_URL, payload2)
        stoken = res.data["token"]
        res = self.client.get(MANAGE_URL)
        self.assertEqual(res.status_code, status.HTTP_401_UNAUTHORIZED)
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)
        res = self.client.get(MANAGE_URL)
        self.assertEqual(res.status_code, status.HTTP_200_OK)
        res = self.client.get(MANAGE_URL)
        self.assertEqual(res.status_code, status.HTTP_200_OK)

        self.client.credentials(HTTP_AUTHORIZATION="Token " + stoken)
        self.client.post(LOGOUT_EVERYONE_URL, {})
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)
        res = self.client.get(MANAGE_URL)
        self.assertEqual(res.status_code, status.HTTP_401_UNAUTHORIZED)
        get_user(email=payload["email"]).delete()
        get_user(email=payload2["email"]).delete()

    def test_logout_everyone_requires_admin(self):
        """Test admin level logout everyone"""
        payload2 = {"email": "super@jaseci.com", "password": "testpass"}
        superuser = create_superuser(**payload2)
        superuser.save()
        payload = {"email": "jscitest_test@jaseci.com", "password": "testpass"}
        user = create_user(**payload)
        user.is_activated = True
        user.save()

        res = self.client.post(TOKEN_URL, payload)
        token = res.data["token"]
        res = self.client.post(TOKEN_URL, payload2)
        # stoken = res.data['token']
        res = self.client.get(MANAGE_URL)
        self.assertEqual(res.status_code, status.HTTP_401_UNAUTHORIZED)
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)
        res = self.client.get(MANAGE_URL)
        self.assertEqual(res.status_code, status.HTTP_200_OK)
        res = self.client.get(MANAGE_URL)
        self.assertEqual(res.status_code, status.HTTP_200_OK)

        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)
        res = self.client.post(LOGOUT_EVERYONE_URL, {})
        self.assertEqual(res.status_code, status.HTTP_403_FORBIDDEN)
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)
        res = self.client.get(MANAGE_URL)
        self.assertEqual(res.status_code, status.HTTP_200_OK)
        get_user(email=payload["email"]).delete()
        get_user(email=payload2["email"]).delete()

    def test_create_token_for_user(self):
        """Test that a token is created for the user"""
        payload = {"email": "jscitest_test@jaseci.com", "password": "testpass"}
        user = create_user(**payload)
        user.is_activated = True
        user.save()
        res = self.client.post(TOKEN_URL, payload)

        # self.assertIsNotNone(get_user(email=payload['email']).hook)

        self.assertIn("token", res.data)
        self.assertEqual(res.status_code, status.HTTP_200_OK)
        get_user(email=payload["email"]).delete()

    def test_create_token_invalid_credentials(self):
        """Test that token is not created if invalid credentials are given"""
        create_user(email="jscitest_test@jaseci.com", password="testpass")
        payload = {"email": "jscitest_test@jaseci.com", "password": "wrong"}
        res = self.client.post(TOKEN_URL, payload)

        self.assertNotIn("token", res.data)
        self.assertEqual(res.status_code, status.HTTP_400_BAD_REQUEST)
        get_user(email=payload["email"]).delete()

    def test_create_token_no_user(self):
        """Test that token is not created if user doens't exist"""
        payload = {"email": "jscitest_test@jaseci.com", "password": "testpass"}
        res = self.client.post(TOKEN_URL, payload)

        self.assertNotIn("token", res.data)
        self.assertEqual(res.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_token_missing_field(self):
        """Test that email and password are required"""
        res = self.client.post(TOKEN_URL, {"email": "one", "password": ""})
        self.assertNotIn("token", res.data)
        self.assertEqual(res.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retrieve_user_unauthorized(self):
        """Test that authentication required for users"""
        res = self.client.get(MANAGE_URL)

        self.assertEqual(res.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_password_reset_endpoint(self):
        """Test that password resent endpoint succeeds for the user"""
        payload = {"email": "jscitest_test@jaseci.com", "password": "testpass"}
        user = create_user(**payload)
        user.is_activated = True
        user.save()

        email_pl = {"email": "jscitest_test@jaseci.com"}

        res = self.client.post(PASSWORD_RESET_URL, email_pl)

        self.assertEqual(res.status_code, status.HTTP_200_OK)
        get_user(email=payload["email"]).delete()

    def test_password_reset_endpoint_invalid_email(self):
        """Test that password resent endpoint succeeds for the user"""
        payload = {"email": "jscitest_test@jaseci.com", "password": "testpass"}
        user = create_user(**payload)
        user.is_activated = True
        user.save()

        email_pl = {"email": "jjscitest_test@jaseci.com"}

        res = self.client.post(PASSWORD_RESET_URL, email_pl)

        self.assertEqual(res.status_code, status.HTTP_400_BAD_REQUEST)
        get_user(email=payload["email"]).delete()


class UserApiPrivateTests(TestCaseHelper, TestCase):
    """Test API requests that are authenticated"""

    def setUp(self):
        super().setUp()
        self.user = create_user(
            email="jscitest_test@jaseci.com",
            password="testpass",
            name="some dude",
        )
        self.client = APIClient()
        self.client.force_authenticate(user=self.user)

    def tearDown(self):
        self.user.delete()
        super().tearDown()

    def test_retrieve_user_success(self):
        """Test retrieving profile for logged in user"""
        res = self.client.get(MANAGE_URL)

        self.assertEqual(res.status_code, status.HTTP_200_OK)
        self.assertEqual(
            res.data,
            {
                "id": self.user.id,
                "name": self.user.name,
                "email": self.user.email,
                "is_activated": self.user.is_activated,
                "is_superuser": self.user.is_superuser,
            },
        )

    def test_post_me_not_allowed(self):
        """Test that POST is not allowed to update user"""
        res = self.client.post(MANAGE_URL, {})

        self.assertEqual(res.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_update_user_profile(self):
        """Test updating the user profile for authenticated user"""
        payload = {"name": "new name", "password": "newpassword123"}

        res = self.client.patch(MANAGE_URL, payload)

        self.user.refresh_from_db()
        self.assertEqual(self.user.name, payload["name"])
        self.assertTrue(self.user.check_password(payload["password"]))
        self.assertEqual(res.status_code, status.HTTP_200_OK)

    def test_update_user_profile_via_put(self):
        """Test updating the user profile for authenticated user"""
        payload = {
            "name": "new name",
            "password": "newpassword123",
            "email": "mars.ninja@gmail.com",
            "is_activated": True,
        }

        res = self.client.put(MANAGE_URL, payload)

        self.user.refresh_from_db()
        self.assertEqual(self.user.name, payload["name"])
        self.assertTrue(self.user.check_password(payload["password"]))
        self.assertEqual(res.status_code, status.HTTP_200_OK)

    def test_manage_user_by_id(self):
        """Test manage user by id"""
        payload = {"email": f"{uuid4()}@jaseci.com", "password": "testpass"}
        user = create_user(**payload)
        user.save()

        self.assertEqual(user.is_activated, False)
        self.assertEqual(user.is_superuser, False)

        payload = {"email": "super@jaseci.com", "password": "testpass"}
        superuser = create_superuser(**payload)
        superuser.save()

        res = self.client.post(TOKEN_URL, payload).data
        self.client.credentials(HTTP_AUTHORIZATION="Token " + res["token"])

        payload = {
            "is_activated": True,
            "is_superuser": True,
        }

        res = self.client.post(f"{MANAGE_URL}{user.id}", payload, format="json")

        self.assertEqual(res.status_code, status.HTTP_200_OK)
        self.assertEqual(res.data, "Update Success!")

        res = self.client.post(f"{MANAGE_URL}{user.id}", payload, format="json")

        self.assertEqual(res.status_code, status.HTTP_200_OK)
        self.assertEqual(res.data, "No changes found!")

        user = get_user(id=user.id)

        self.assertEqual(user.is_activated, True)
        self.assertEqual(user.is_superuser, True)



from django.urls import path
from .views import AbstractJacAPIView
from .views import AbstractAdminJacAPIView, AbstractPublicJacAPIView
from jaseci.element.element import Element
from jaseci_serv.base.models import SuperMaster
from jaseci.utils.utils import copy_func
from inspect import signature
import uuid
import json


def rest_api_auto_doc(endpoint: str, fsig: signature):
    """
    Automatically return string for REST Interface documentation
    Parameters include endpoint uri and signature
    """
    doc = f"**REST Endpoint**: {endpoint}\n"
    params = []
    json_samp = {}

    for i in fsig.parameters.keys():
        if i == "self":
            continue
        p_name = i
        p_type = fsig.parameters[i].annotation
        if issubclass(p_type, Element):
            params.append(f"> {p_name}: UUID pointing to {p_type.__name__} object\n")
            json_samp[p_name] = uuid.uuid4().urn
        else:
            params.append(f"> {p_name}: type {p_type.__name__}\n")
            json_samp[p_name] = p_type()
            if p_type == str:
                json_samp[p_name] = "some string"
    if params:
        doc += "\n**Parameters**:\n"
        for i in params:
            doc += i
        doc += "\n**JSON Example**\n"
        doc += f"```javascript\n{json.dumps(json_samp, indent=4)}\n```\n"
    return doc


def rest_api_build_methods(api, view_cls, fname, apidocstr, func_sig):
    gen_cls = type(fname, (view_cls,), {})
    allowed_methods = (
        getattr(gen_cls, "http_method_names")
        if api["allowed_methods"] is None
        else api["allowed_methods"]
    )
    setattr(gen_cls, "http_method_names", allowed_methods)
    for method in allowed_methods:
        setattr(gen_cls, method, copy_func(getattr(gen_cls, method)))
        getattr(gen_cls, method).__doc__ = (
            api["doc"] + "\n\n" + rest_api_auto_doc(apidocstr, func_sig)
        )
    return gen_cls


generated_urls = []


def generate_apis(api_list, view_cls, dir_head):
    """
    Auto generates Django APIs based on core interface
    """
    for i in api_list:
        fname = "_".join(i["groups"])
        apidocstr = f"{dir_head}/{fname}"

        globals()[fname] = rest_api_build_methods(
            i, view_cls, fname, apidocstr, i["sig"]
        )

        global generated_urls
        url_args = ""
        for j in i["url_args"]:
            url_args += f"/<str:{j}>"
        generated_urls.append(
            path(apidocstr + url_args, globals()[fname].as_view(), name=fname)
        )


generate_apis(SuperMaster._public_api, AbstractPublicJacAPIView, "js_public")
generate_apis(SuperMaster._private_api, AbstractJacAPIView, "js")
generate_apis(SuperMaster._admin_api, AbstractAdminJacAPIView, "js_admin")



from django.conf import settings
from django.core.management.base import BaseCommand
import subprocess


class Command(BaseCommand):
    """
    Create postgres database
    https://www.postgresql.org/docs/current/app-createdb.html
    """

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument("alias", nargs="?", default="default")

    def handle(self, *args, **options):
        alias = options.get("alias")

        db_settings = settings.DATABASES[alias]
        args = ["createdb"]
        if "USER" in db_settings:
            args += ["-U", db_settings["USER"]]
        if "HOST" in db_settings:
            args += ["-h", db_settings["HOST"]]
        if "PORT" in db_settings:
            args += ["-p", str(db_settings["PORT"])]
        args.append(db_settings["NAME"])
        subprocess.check_call(args)



from jaseci.svc import RedisService as Rs
from jaseci_serv.jaseci_serv.settings import REDIS_CONFIG


class RedisService(Rs):
    def get_config(self, hook) -> dict:
        return hook.build_config("REDIS_CONFIG", REDIS_CONFIG)



from jaseci.svc import TaskService as Ts
from jaseci_serv.jaseci_serv.settings import TASK_CONFIG


#################################################
#                 TASK APP ORM                  #
#################################################


class TaskService(Ts):
    def get_config(self, hook) -> dict:
        return hook.build_config("TASK_CONFIG", TASK_CONFIG)



from jaseci.svc import ServiceState
from .redis import RedisService
from .task import TaskService
from .mail import MailService
from .meta import MetaService

__all__ = [
    "ServiceState",
    "RedisService",
    "TaskService",
    "MailService",
    "MetaService",
]



from django.contrib.auth import get_user_model
from django.urls import reverse

from rest_framework.test import APIClient
from jaseci.utils.utils import TestCaseHelper
import jaseci.actions.live_actions as lact
from django.test import TestCase
import uuid
import base64
import os


class TestLLWall(TestCaseHelper, TestCase):
    """Test the authorized user node API"""

    def setUp(self):
        super().setUp()
        self.user = get_user_model().objects.create_user(
            "JSCITfdfdEST_test@jaseci.com", "password"
        )
        self.master = self.user.get_master()
        self.client = APIClient()
        self.client.force_authenticate(self.user)
        ll_loc = os.path.dirname(__file__) + "/ll_wall.jac"
        ll_file = base64.b64encode(open(ll_loc).read().encode()).decode()
        payload = {
            "op": "sentinel_register",
            "name": "Something",
            "code": ll_file,
            "encoded": True,
        }
        lact.load_local_actions("jaseci_serv/jac_api/tests/infer.py")
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.snt = self.master._h.get_obj(
            self.master.jid, uuid.UUID(res.data[0]["jid"])
        )
        self.gph = self.master._h.get_obj(
            self.master.jid, uuid.UUID(res.data[1]["jid"])
        )

    def tearDown(self):
        super().tearDown()

    def run_walker(self, w_name, ctx, prime=None):
        """Helper to make calls to execute walkers"""
        if not prime:
            payload = {
                "snt": self.snt.id.urn,
                "name": w_name,
                "nd": self.gph.id.urn,
                "ctx": ctx,
            }
        else:
            payload = {"snt": self.snt.id.urn, "name": w_name, "nd": prime, "ctx": ctx}
        res = self.client.post(reverse("jac_api:walker_run"), payload, format="json")
        return res.data

    def graph_node_set(self, nd_id, ctx):
        """Helper to set node context"""
        payload = {"snt": self.snt.id.urn, "nd": nd_id, "ctx": ctx}
        res = self.client.post(
            reverse("jac_api:graph_node_set"), payload, format="json"
        )
        return res.data

    def test_ll_wall_get_gen_day(self):
        """Test get_gen_day walker response time after cerify day"""
        num_workettes = 3
        # generate random day workettes
        self.run_walker(
            "gen_day_workettes", {"date": "2021-07-12", "num_workettes": num_workettes}
        )
        data = self.run_walker("get_latest_day", {"show_report": 1})["report"]

        day_id = data[0][1]["jid"]
        day_date = data[0][1]["context"]["day"]
        day_note = data[0][1]["context"]["note"]

        data = self.run_walker("get_workettes_deep", {"show_report": 1}, prime=day_id)[
            "report"
        ]
        self.assertEqual(len(data[0]), num_workettes)

        # certify day, should return day highlights
        data = self.run_walker(
            "set_day_highlight",
            {
                "highlight_items": [
                    {
                        "id": data[0][0],
                        "type": "Most proud accomplishment",
                        "color": "#464ff6",
                        "icon": "0x1F3C6",
                    },
                    {
                        "id": data[0][1],
                        "type": "Made You Happiest",
                        "color": "#6e30dd",
                        "icon": "0x1F604",
                    },
                    {
                        "id": data[0][2],
                        "type": "Required the Most Work",
                        "color": "#b926df",
                        "icon": "0x1F4AA",
                    },
                ]
            },
            prime=day_id,
        )["report"]
        self.assertEqual(data[0][0][0][0][0]["jid"], day_id)
        self.assertEqual(data[0][0][0][0][1], day_date)
        self.assertEqual(data[0][0][0][1], day_note)

        # data[0][0][0][2] is the highlight items
        self.assertEqual(len(data[0][0][0][2]), 3)



import json
from base64 import b64encode
from io import BytesIO
from tempfile import _TemporaryFileWrapper
from time import time

from knox.auth import TokenAuthentication
from rest_framework.permissions import AllowAny, IsAdminUser, IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from jaseci.element.element import Element
from jaseci.utils.utils import logger, ColCodes as Cc
from jaseci_serv.base.models import Master as ServMaster
from jaseci_serv.svc import MetaService


class JResponse(Response):
    def __init__(self, master, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hook = master._h
        for i in self.hook.save_obj_list:
            self.hook.commit_obj_to_cache(i)

    def close(self):
        super(JResponse, self).close()
        # Commit db changes after response to user
        self.hook.commit(True)


class AbstractJacAPIView(APIView):
    """
    The builder set of Jaseci APIs
    """

    http_method_names = ["post"]
    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    def get(self, request, **kwargs):
        """
        General GET function that parses api signature to load parms
        SuperSmart GET - can read signatures of master and process
        bodies accordingly
        """
        self.proc_request(request, **kwargs)
        api_result = self.caller.general_interface_to_api(self.cmd, type(self).__name__)
        self.log_request_stats()
        return self.issue_response(api_result)

    def post(self, request, **kwargs):
        """
        General post function that parses api signature to load parms
        SuperSmart Post - can read signatures of master and process
        bodies accordingly
        """
        self.proc_request(request, **kwargs)
        api_result = self.caller.general_interface_to_api(self.cmd, type(self).__name__)
        self.log_request_stats()
        return self.issue_response(api_result)

    def log_request_stats(self):
        """Api call preamble"""
        tot_time = time() - self.start_time
        save_count = 0
        if isinstance(self.caller, Element):
            save_count = len(self.caller._h.save_obj_list)
        logger.info(
            str(
                f"API call to {Cc.TG}{type(self).__name__}{Cc.EC}"
                f" completed in {Cc.TY}{tot_time:.3f} seconds{Cc.EC}"
                f" saving {Cc.TY}{save_count}{Cc.EC} objects."
            )
        )

    def proc_prime_ctx(self, request, req_data):
        try:
            if "ctx" in request.FILES:
                ctx = request.FILES.pop("ctx")[0]
                if ctx.content_type == "application/json":
                    req_data["ctx"] = json.loads(ctx.read().decode("utf-8"))
            elif "ctx" in req_data and type(req_data["ctx"]) is not dict:
                req_data["ctx"] = json.loads(req_data["ctx"])
        except ValueError:
            logger.error("Invalid ctx format! Ignoring ctx parsing!")

    def proc_file_ctx(self, request, req_data):
        for key in request.FILES:

            req_data.pop(key)
            file_ref = (
                req_data["ctx"]
                if "ctx" in req_data and type(req_data["ctx"]) is dict
                else req_data
            )
            file_ref[key] = []

            for file in request.FILES.getlist(key):

                file_type = type(file.file)
                if file_type is BytesIO:
                    file_base64 = b64encode(file.file.getvalue()).decode("utf-8")
                elif file_type is _TemporaryFileWrapper:
                    file_base64 = b64encode(file.file.read()).decode("utf-8")

                if "file_base64" in vars():
                    file_ref[key].append(
                        {
                            "name": file.name,
                            "base64": file_base64,
                            "content-type": file.content_type,
                        }
                    )

    def proc_request_ctx(self, request, req_data):
        req_query = request.GET.dict()
        req_data.update(
            {
                "_req_ctx": {
                    "method": request.method,
                    "headers": dict(request.headers),
                    "query": req_query,
                    "body": req_data.copy(),
                }
            }
        )
        req_data.update(req_query)

    def proc_request(self, request, **kwargs):
        """Parse request to field set"""
        pl_peek = str(dict(request.data))[:256]
        logger.info(str(f"Incoming call to {type(self).__name__} with {pl_peek}"))
        self.start_time = time()

        req_data = (
            request.data.dict() if type(request.data) is not dict else request.data
        )

        req_data.update(kwargs)

        self.proc_prime_ctx(request, req_data)
        self.proc_file_ctx(request, req_data)
        self.proc_request_ctx(request, req_data)

        self.cmd = req_data
        self.set_caller(request)
        self.res = "Not valid interaction!"

    def set_caller(self, request):
        """Assigns the calling api interface obj"""
        self.caller = request.user.get_master()

    def pluck_status_code(self, api_result):
        """Extracts status code out of payload"""
        status = 200
        if isinstance(api_result, dict) and "status_code" in api_result:
            status = api_result["status_code"]
        return status

    def issue_response(self, api_result):
        """Issue response from call"""
        # self.caller._h.commit()
        # return Response(api_result)
        # for i in self.caller._h.save_obj_list:
        #     self.caller._h.commit_obj_to_redis(i)
        status = self.pluck_status_code(api_result)
        if isinstance(api_result, dict) and "report_custom" in api_result.keys():
            api_result = api_result["report_custom"]
        return JResponse(self.caller, api_result, status=status)


class AbstractAdminJacAPIView(AbstractJacAPIView):
    """
    The abstract base for Jaseci Admin APIs
    """

    http_method_names = ["post"]
    permission_classes = (IsAuthenticated, IsAdminUser)


class AbstractPublicJacAPIView(AbstractJacAPIView):
    """
    The abstract base for Jaseci Admin APIs
    """

    http_method_names = ["post"]
    permission_classes = (AllowAny,)

    def get(self, request, **kwargs):
        """
        Public GET function that parses api signature to load parms
        SuperSmart GET - can read signatures of master and process
        bodies accordingly
        """
        self.proc_request(request, **kwargs)

        api_result = self.caller.public_interface_to_api(self.cmd, type(self).__name__)
        self.log_request_stats()
        return self.issue_response(api_result)

    def post(self, request, **kwargs):
        """
        Public post function that parses api signature to load parms
        SuperSmart Post - can read signatures of master and process
        bodies accordingly
        """
        self.proc_request(request, **kwargs)

        api_result = self.caller.public_interface_to_api(self.cmd, type(self).__name__)
        self.log_request_stats()
        return self.issue_response(api_result)

    def set_caller(self, request):
        """Assigns the calling api interface obj"""
        self.caller = ServMaster(
            h=MetaService().hook(),
            persist=False,
        )

    def issue_response(self, api_result):
        """Issue response from call"""
        # If committer set, results should be saved back
        status = self.pluck_status_code(api_result)

        if isinstance(api_result, dict) and "report_custom" in api_result.keys():
            api_result = api_result["report_custom"]

        if self.caller._pub_committer:
            return JResponse(self.caller._pub_committer, api_result, status=status)
        else:
            return Response(api_result, status=status)



from django.urls import path, include
from rest_framework.routers import DefaultRouter

from jaseci_serv.obj_api import views

router = DefaultRouter()
router.register("obj", views.ObjectViewSet)
router.register("global", views.ConfigViewSet)

app_name = "obj_api"

urlpatterns = [path("", include(router.urls))]



"""
Django settings for jaseci project.

Generated by 'django-admin startproject' using Django 3.0.3.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.0/ref/settings/
"""

import os
import sys

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEMPLATE_DIR = os.path.join(BASE_DIR, "ui/templates")

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "@%ad_kyis62uf7qf^w9kv(8$db4)%c$nnnjk^us=s@-gj*)aal"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["*"]

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "rest_framework",
    "knox",
    "django_rest_passwordreset",
    "drf_yasg",
    "jaseci_serv.base",
    "jaseci_serv.user_api",
    "jaseci_serv.obj_api",
    "jaseci_serv.jac_api",
    "corsheaders",
    "django_celery_results",
    "django_celery_beat",
]

MIDDLEWARE = [
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

CORS_ORIGIN_ALLOW_ALL = True

ROOT_URLCONF = "jaseci_serv.jaseci_serv.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [TEMPLATE_DIR],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "jaseci_serv.jaseci_serv.wsgi.application"

# Database
# https://docs.djangoproject.com/en/3.0/ref/settings/#databases

if "RDS_HOSTNAME" in os.environ:
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.mysql",
            "NAME": os.environ["RDS_DB_NAME"],
            "USER": os.environ["RDS_USERNAME"],
            "PASSWORD": os.environ["RDS_PASSWORD"],
            "HOST": os.environ["RDS_HOSTNAME"],
            "PORT": os.environ["RDS_PORT"],
        }
    }

elif "POSTGRES_HOST" in os.environ:
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql",
            "HOST": os.environ.get("POSTGRES_HOST"),
            "NAME": os.environ.get("POSTGRES_USER"),
            "USER": os.environ.get("POSTGRES_USER"),
            "PASSWORD": os.environ.get("POSTGRES_PASSWORD"),
            "PORT": os.getenv("POSTGRES_PORT", 5432),
        }
    }

else:
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": "mydatabase",
            "TEST": {"NAME": "test"},
        }
    }

RUN_SVCS = "test" in sys.argv or "runserver" in sys.argv

# REDIS
REDIS_CONFIG = {
    "enabled": True,
    "quiet": False,
    "host": os.getenv("REDIS_HOST", "localhost"),
    "port": os.getenv("REDIS_PORT", "6379"),
    "db": os.getenv("REDIS_DB", "1"),
}

# TASK_HOOK
TASK_CONFIG = {
    "enabled": True,
    "quiet": False,
    "broker_url": f"redis://{REDIS_CONFIG.get('host')}:{REDIS_CONFIG['port']}/{REDIS_CONFIG['db']}",
    "beat_scheduler": "django_celery_beat.schedulers:DatabaseScheduler",
    "result_backend": "django-db",
    "task_track_started": True,
    "broker_connection_retry_on_startup": True,
}

# EMAIL_HOOK
MAIL_CONFIG = {
    "enabled": True,
    "quiet": False,
    "version": 1,
    "tls": True,
    "host": "smtp.gmail.com",
    "port": 587,
    "sender": "Jaseci Admin<boyong@jaseci.org>",
    "user": "jaseci.dev@gmail.com",
    "pass": "yrtviyrdzmzdpjxg",
    "backend": "smtp",
    "templates": {
        "activation_subj": "Please activate your account!",
        "activation_body": "Thank you for creating an account!\n\n"
        "Activation Code: {{code}}\n"
        "Please click below to activate:\n{{link}}",
        "activation_html_body": "Thank you for creating an account!<br><br>"
        "Activation Code: {{code}}<br>"
        "Please click below to activate:<br>"
        "{{link}}",
        "resetpass_subj": "Password Reset for Jaseci Account",
        "resetpass_body": "Your Jaseci password reset token is: {{token}}",
        "resetpass_html_body": "Your Jaseci password reset" "token is: {{token}}",
    },
    "migrate": False,
}

if "test" in sys.argv or "test_coverage" in sys.argv:
    MAIL_CONFIG["backend"] = "locmem"
    TASK_CONFIG["task_always_eager"] = True
    TASK_CONFIG["task_store_eager_result"] = True
    TASK_CONFIG["beat_scheduler"] = "celery.beat:PersistentScheduler"

# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation."
        + "UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation." + "MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation." + "CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation." + "NumericPasswordValidator",
    },
]

# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = False

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/

STATIC_URL = "/static/"
STATICFILES_DIRS = [os.path.join(BASE_DIR, "static")]

# Sets global default user model to be custom
AUTH_USER_MODEL = "base.User"

REST_FRAMEWORK = {
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "PAGE_SIZE": 100,
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "knox.auth.TokenAuthentication",
    ],
}

REST_KNOX = {
    "TOKEN_TTL": None,
}

LOGIN_URL = "/ui/login"
LOGIN_REDIRECT_URL = "/ui"
LOGOUT_REDIRECT_URL = "/ui"

# Configs to Manage by Jaseci
JASECI_CONFIGS = [
    "EMAIL_BACKEND",
    "EMAIL_USE_TLS",
    "EMAIL_HOST",
    "EMAIL_HOST_USER",
    "EMAIL_HOST_PASSWORD",
    "EMAIL_PORT",
    "EMAIL_DEFAULT_FROM",
    "EMAIL_ACTIVATION_SUBJ",
    "EMAIL_ACTIVATION_BODY",
    "EMAIL_ACTIVATION_HTML_BODY",
    "EMAIL_RESETPASS_SUBJ",
    "EMAIL_RESETPASS_BODY",
    "EMAIL_RESETPASS_HTML_BODY",
    "LOGIN_TOKEN_TTL_HOURS",
]

DJANGO_CELERY_BEAT_TZ_AWARE = False
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"



from django.core import mail

from jaseci.svc import MailService as Ms
from jaseci.svc.mail import Mailer as Em
from jaseci_serv.jaseci_serv.settings import MAIL_CONFIG


#################################################
#                 EMAIL APP ORM                 #
#################################################


class MailService(Ms):
    def connect(self, configs):
        user = configs.get("user")
        backend = configs.get("backend", "smtp")
        sender = configs.get("sender", user)

        server = mail.get_connection(
            backend=f"django.core.mail.backends.{backend}.EmailBackend",
            host=configs.get("host"),
            port=configs.get("port"),
            username=user,
            password=configs.get("pass"),
            use_tls=configs.get("tls"),
        )
        return Mailer(server, sender, configs["templates"])

    def get_config(self, hook) -> dict:
        return hook.build_config("MAIL_CONFIG", MAIL_CONFIG)


# ----------------------------------------------- #


####################################################
#                   EMAIL CONFIG                   #
####################################################


class Mailer(Em):
    def __init__(self, server, sender, templates):
        super().__init__(server, sender)

        self.activ_subj = templates["activation_subj"]
        self.activ_body = templates["activation_body"]
        if "{{code}}" not in self.activ_body and "{{link}}" not in self.activ_body:
            raise Exception("{{code/link}} must be present in email template")

        self.activ_html = templates["activation_html_body"]
        if "{{code}}" not in self.activ_html and "{{link}}" not in self.activ_html:
            raise Exception("{{code/link}} must be present in email template")

        self.reset_subj = templates["resetpass_subj"]
        self.reset_body = templates["resetpass_body"]
        if "{{token}}" not in self.reset_body:
            raise Exception("{{token}} must be present in email template")

        self.reset_html = templates["resetpass_html_body"]
        if "{{token}}" not in self.reset_html:
            raise Exception("{{token}} must be present in email template")

    def send_activation_email(self, email, code, link):
        """Apply relevant parameters to loaded templates"""
        body = self.activ_body.replace("{{code}}", code).replace("{{link}}", link)
        html = self.activ_html.replace("{{code}}", code).replace("{{link}}", link)

        msg = mail.EmailMultiAlternatives(
            subject=self.activ_subj,
            body=body,
            from_email=self.sender,
            to=[email],
            connection=self.server,
        )

        msg.attach_alternative(html, "text/html")
        msg.send(fail_silently=False)

    def send_reset_email(self, email, token):
        """Apply relevant parameters to loaded templates"""
        body = self.reset_body.replace("{{token}}", token)
        html = self.reset_html.replace("{{token}}", token)

        msg = mail.EmailMultiAlternatives(
            subject=self.reset_subj,
            body=body,
            from_email=self.sender,
            to=[email],
            connection=self.server,
        )
        msg.attach_alternative(html, "text/html")
        msg.send(fail_silently=False)

    def send_custom_email(
        self,
        sender: str = None,
        recipients: list = [],
        subject: str = "Jaseci Email",
        body: tuple = ("", ""),
    ):
        msg = mail.EmailMultiAlternatives(
            subject=subject,
            body=body[0],
            from_email=self.sender if sender is None else sender,
            to=recipients,
            connection=self.server,
        )
        msg.attach_alternative(body[1], "text/html")
        msg.send(fail_silently=False)

    def terminate(self):
        self.server.close()



import base64

from django.contrib.auth import authenticate, get_user_model
from django.dispatch import receiver
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django_rest_passwordreset.signals import reset_password_token_created
from rest_framework import serializers

from jaseci.svc.mail import MAIL_ERR_MSG
from jaseci.utils.utils import logger
from jaseci_serv.svc import MailService


def send_activation_email(request, email):
    """Construct activation email body"""
    code = base64.b64encode(email.encode()).decode()
    link = request.build_absolute_uri(
        reverse("user_api:activate", kwargs={"code": code})
    )
    ma = MailService()
    if ma.is_running():
        ma.app.send_activation_email(email, code, link)
    else:
        logger.warning(MAIL_ERR_MSG)


@receiver(reset_password_token_created)
def password_reset_token_created(
    sender, instance, reset_password_token, *args, **kwargs
):
    ma = MailService()
    if ma.is_running():
        ma.app.send_reset_email(
            reset_password_token.user.email, reset_password_token.key
        )
    else:
        logger.warning(MAIL_ERR_MSG)


class UserSerializer(serializers.ModelSerializer):
    """Serializer for users object"""

    class Meta:
        model = get_user_model()
        fields = ("id", "email", "password", "name", "is_activated", "is_superuser")
        extra_kwargs = {"password": {"write_only": True, "min_length": 8}}

    def create(self, validated_data):
        """Create a new user with encrypted password and return it"""
        return get_user_model().objects.create_user(**validated_data)

    def update(self, instance, validated_data):
        """Update user, setting the password if needed"""
        password = validated_data.pop("password", None)
        user = super().update(instance, validated_data)

        if password:
            user.set_password(password)
            user.save()

        return user


class SuperUserSerializer(UserSerializer):
    """Serializer for users object"""

    def create(self, validated_data):
        """Create a new user with encrypted password and return it"""
        return get_user_model().objects.create_superuser(**validated_data)


class AuthTokenSerializer(serializers.Serializer):
    """Serializer for the user authentication object"""

    email = serializers.CharField()
    password = serializers.CharField(
        style={"input_type": "password"}, trim_whitespace=False
    )

    def validate(self, attrs):
        """
        Validate and authenticate the user
        """
        email = attrs.get("email").lower()
        password = attrs.get("password")

        user = authenticate(
            request=self.context.get("request"), username=email, password=password
        )
        if not user:
            msg = _("Unable to authenticate with provided credentials")
            raise serializers.ValidationError(msg, code="authorization")
        elif not user.is_activated:
            msg = _(
                "User not activated. Resending activation email.\n"
                + "Please check your email."
            )
            send_activation_email(self.context.get("request"), email)
            raise serializers.ValidationError(msg)

        attrs["user"] = user
        return attrs



from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework.test import APIClient
from jaseci.utils.utils import TestCaseHelper
import jaseci.actions.live_actions as lact
from django.test import TestCase
import uuid
import base64
import datetime
import random
import os


class TestLL(TestCaseHelper, TestCase):
    """Test the authorized user node API"""

    def setUp(self):
        super().setUp()
        lact.load_local_actions(os.path.dirname(__file__) + "/infer.py")
        self.user = get_user_model().objects.create_user(
            "JSCITfdfdEST_test@jaseci.com", "password"
        )
        self.client = APIClient()
        self.client.force_authenticate(self.user)
        ll_loc = os.path.dirname(__file__) + "/ll.jac"
        ll_file = base64.b64encode(open(ll_loc).read().encode()).decode()
        payload = {
            "op": "sentinel_register",
            "name": "Something",
            "code": ll_file,
            "encoded": True,
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {"op": "global_sentinel_set"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.user = get_user_model().objects.create_user(
            "J2SCITfdfdEST_test@jaseci.com", "password"
        )
        self.client = APIClient()
        self.client.force_authenticate(self.user)
        self.master = self.user.get_master()
        payload = {"op": "sentinel_active_global"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_create"}
        self.snt = self.master._h.get_obj(
            self.master.jid, uuid.UUID(res.data["sentinel"]["jid"])
        )
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.gph = self.master._h.get_obj(self.master.jid, uuid.UUID(res.data["jid"]))

    def tearDown(self):
        super().tearDown()

    def run_walker(self, w_name, ctx, prime=None):
        """Helper to make calls to execute walkers"""
        if not prime:
            payload = {
                "snt": self.snt.id.urn,
                "name": w_name,
                "nd": self.gph.id.urn,
                "ctx": ctx,
            }
        else:
            payload = {"snt": self.snt.id.urn, "name": w_name, "nd": prime, "ctx": ctx}
        res = self.client.post(reverse("jac_api:walker_run"), payload, format="json")
        return res.data

    def graph_node_set(self, nd_id, ctx):
        """Helper to set node context"""
        payload = {"snt": self.snt.id.urn, "nd": nd_id, "ctx": ctx}
        self.client.post(reverse("jac_api:graph_node_set"), payload, format="json")

    def test_ll_today_new(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker("get_gen_day", {})["report"]
        self.assertEqual(data[0][1]["name"], "day")
        jid = data[0][1]["jid"]
        data = self.run_walker("get_gen_day", {})["report"]
        self.assertEqual(data[0][1]["jid"], jid)

    def test_ll_create_get_workette(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker("get_gen_day", {})["report"]
        self.assertEqual(data[0][1]["name"], "day")
        jid = data[0][1]["jid"]
        data = self.run_walker("create_workette", {}, prime=jid)["report"]
        self.assertEqual(data[0]["name"], "workette")
        data = self.run_walker("create_workette", {}, prime=jid)["report"]
        self.assertEqual(data[0]["name"], "workette")
        data = self.run_walker("get_workettes", {}, prime=jid)["report"]
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]["name"], "workette")
        self.assertEqual(data[1]["name"], "workette")
        wjid = data[0]["jid"]
        data = self.run_walker("create_workette", {}, prime=wjid)["report"]
        data = self.run_walker("create_workette", {}, prime=wjid)["report"]
        data = self.run_walker("create_workette", {}, prime=wjid)["report"]
        data = self.run_walker("get_workettes", {}, prime=wjid)["report"]
        self.assertEqual(len(data), 3)
        data = self.run_walker("get_workettes", {}, prime=jid)["report"]
        self.assertEqual(len(data), 2)
        data = self.run_walker("get_workettes_deep", {}, prime=jid)["report"]
        self.assertEqual(len(data), 5)

    def test_ll_get_date(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker("get_gen_day", {})
        data = self.run_walker("gen_rand_life", {})
        data = self.run_walker("get_latest_day", {"show_report": 1})["report"]
        wjid = data[0][1]["jid"]
        data = self.run_walker("get_workettes", {}, prime=wjid)["report"]
        self.assertEqual(len(data), 0)
        data = self.run_walker(
            "get_latest_day", {"before_date": "2020-03-01", "show_report": 1}
        )["report"]
        wjid = data[0][1]["jid"]
        data = self.run_walker("get_workettes", {}, prime=wjid)["report"]
        self.assertGreater(len(data), 1)

    def test_ll_carry_forward_simple(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker("gen_rand_life", {})
        data = self.run_walker("get_gen_day", {})
        wjid = data["report"][0][1]["jid"]
        data = self.run_walker("get_workettes", {}, prime=wjid)
        self.assertGreater(len(data["report"]), 1)

    def test_ll_delete_workette(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker("get_gen_day", {})["report"]
        self.assertEqual(data[0][1]["name"], "day")
        jid = data[0][1]["jid"]
        data = self.run_walker("create_workette", {}, prime=jid)["report"]
        self.assertEqual(data[0]["name"], "workette")
        data = self.run_walker("create_workette", {}, prime=jid)["report"]
        self.assertEqual(data[0]["name"], "workette")
        data = self.run_walker("get_workettes", {}, prime=jid)["report"]
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]["name"], "workette")
        self.assertEqual(data[1]["name"], "workette")
        wjid = data[0]["jid"]
        data = self.run_walker("create_workette", {}, prime=wjid)
        data = self.run_walker("create_workette", {}, prime=wjid)
        data = self.run_walker("create_workette", {}, prime=wjid)
        data = self.run_walker("get_workettes", {}, prime=wjid)["report"]
        self.assertEqual(len(data), 3)
        data = self.run_walker("delete_workette", {}, prime=wjid)
        data = self.run_walker("get_workettes", {}, prime=jid)["report"]
        self.assertEqual(len(data), 1)
        data = self.run_walker("get_workettes_deep", {}, prime=jid)["report"]
        self.assertEqual(len(data), 1)

    def test_ll_delete_workette_no_leaks(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker("get_gen_day", {})["report"]
        self.assertEqual(data[0][1]["name"], "day")
        jid = data[0][1]["jid"]
        data = self.run_walker("create_workette", {}, prime=jid)["report"]
        self.assertEqual(data[0]["name"], "workette")
        len_before = self.master._h.get_object_distribution()
        data = self.run_walker("create_workette", {}, prime=jid)["report"]
        self.assertEqual(data[0]["name"], "workette")
        data = self.run_walker("get_workettes", {}, prime=jid)["report"]
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]["name"], "workette")
        self.assertEqual(data[1]["name"], "workette")
        wjid = data[1]["jid"]
        data = self.run_walker("create_workette", {}, prime=wjid)["report"]
        exjid = data[0]["jid"]
        data = self.run_walker("create_workette", {}, prime=wjid)
        data = self.run_walker("create_workette", {}, prime=wjid)
        data = self.run_walker("get_workettes", {}, prime=wjid)["report"]
        self.assertEqual(len(data), 3)
        data = self.run_walker("delete_workette", {}, prime=wjid)["report"]
        self.assertNotIn(uuid.UUID(exjid).urn, self.master._h.mem.keys())
        self.assertIn(uuid.UUID(jid).urn, self.master._h.mem.keys())
        data = self.run_walker("get_workettes", {}, prime=jid)["report"]
        self.assertEqual(len(data), 1)
        data = self.run_walker("get_workettes_deep", {}, prime=jid)["report"]
        self.assertEqual(len(data), 1)
        len_after = self.master._h.get_object_distribution()
        self.assertEqual(len_before, len_after)

    def test_due_soon(self):
        """Test generating a list of suggested focus items for a given day"""
        self.run_walker("gen_rand_life", {})
        self.run_walker("get_gen_day", {})
        data = self.run_walker("get_latest_day", {"show_report": 1})
        w_id = data["report"][0][1]["jid"]

        today_date = datetime.datetime.today()
        due_item = False
        for wkt in data["report"]:
            if wkt[1]["name"] == "workette":
                if random.choice([0, 1]) or not due_item:
                    due_item = True
                    due = random.randint(-3, 3)
                    due_date = today_date + datetime.timedelta(days=due)
                    due_date = due_date.replace(
                        hour=0, minute=0, second=0, microsecond=0
                    )
                    wkt[1]["context"]["date"] = due_date.isoformat()
                    self.graph_node_set(wkt[1]["jid"], wkt[1]["context"])

        data = self.run_walker(
            "get_due_soon", {"soon": 4, "show_report": 1}, prime=w_id
        )
        self.assertTrue(len(data["report"]) > 0)

    def test_days_in_backlog(self):
        """Test getting the number of days an item has been in the backlog"""
        self.run_walker("gen_rand_life", {})
        self.run_walker("get_gen_day", {})
        data = self.run_walker("get_latest_day", {"show_report": 1})
        w_id = data["report"][0][1]["jid"]
        workettes = self.run_walker("get_workettes", {}, prime=w_id)["report"]
        for wkt in workettes:
            res = self.run_walker(
                "days_in_backlog", {"show_report": 1}, prime=wkt["jid"]
            )
            self.assertIs(type(res["report"][0]), int)

    def test_get_long_active_items(self):
        """Test getting items been in backlog for long"""
        self.run_walker("gen_rand_life", {})
        self.run_walker("get_gen_day", {})
        data = self.run_walker("get_latest_day", {"show_report": 1})
        w_id = data["report"][0][1]["jid"]
        result = self.run_walker(
            "get_long_active_items", {"show_report": 1, "long_days": 1}, prime=w_id
        )
        self.assertTrue(len(result) > 0)

    def test_get_suggested_focus(self):
        self.run_walker("gen_rand_life", {})
        self.run_walker("get_gen_day", {})
        data = self.run_walker("get_latest_day", {"show_report": 1})
        w_id = data["report"][0][1]["jid"]
        result = self.run_walker(
            "get_suggested_focus", {"max_items": 5, "long_days": 1}, prime=w_id
        )
        self.assertTrue(len(result) > 0)



from django.contrib.auth import get_user_model
from django.test import TestCase

from jaseci.element import element
from jaseci.element.super_master import SuperMaster
from jaseci.utils.utils import TestCaseHelper
from jaseci_serv.base import models
from jaseci_serv.obj_api.views import JaseciObjectSerializer
from jaseci_serv.svc import MetaService


def sample_user(email="JSCITEST_user@jaseci.com", password="whatever"):
    """Create a sample user for testing"""
    if get_user_model().objects.filter(email=email).exists():
        return get_user_model().objects.get(email=email)
    return get_user_model().objects.create_user(email, password)


class ModelTests(TestCaseHelper, TestCase):
    def setUp(self):
        super().setUp()
        self.meta = MetaService()

    def tearDown(self):
        super().tearDown()

    def test_users_creation_and_deletion(self):
        """
        Tests that users are created based on valid emails and
        objects are deleted with user
        """
        email = "JSCITEST_blah@BlaB.com"
        password = "passW123"
        num_nodes = models.JaseciObject.objects.count()
        user = get_user_model().objects.create_user(email=email, password=password)
        # This tests that nodes are created/destroyed along side users
        self.assertEqual(num_nodes + 1, models.JaseciObject.objects.count())

        self.assertEqual(user.email.split("@")[1], email.split("@")[1].lower())
        self.assertTrue(user.check_password(password))
        user.delete()
        self.assertFalse(get_user_model().objects.filter(id=user.id).exists())
        self.assertEqual(num_nodes, models.JaseciObject.objects.count())

    def test_create_super_user(self):
        """Tests that superuser is created and has the right permissions"""
        user = get_user_model().objects.create_superuser(
            email="JSCITEST_super@User.com", password="135jj"
        )

        self.assertIsInstance(user.get_master(), SuperMaster)
        user.delete()
        self.assertFalse(get_user_model().objects.filter(id=user.id).exists())

    def test_jaseci_obj_accessl_has_relevant_fields(self):
        """Test that Jaseci ORM models has all element class fields"""
        element_obj = element.Element(m_id="anon", h=self.meta.hook())
        orm_obj = models.JaseciObject()
        for a in vars(element_obj).keys():
            if not a.startswith("_") and not callable(getattr(element_obj, a)):
                self.assertIn(a, dir(orm_obj))

    def test_jaseci_json_has_relevant_fields(self):
        """Test that Jaseci ORM models has all element class fields"""
        element_obj = element.Element(m_id="anon", h=self.meta.hook())
        for a in vars(element_obj).keys():
            if not a.startswith("_") and not callable(getattr(element_obj, a)):
                self.assertIn(a, JaseciObjectSerializer.Meta.fields)

    def test_jaseci_obj_accessl_create_delete(self):
        """Test that we can create and delete jaseci object models"""
        orm_obj = models.JaseciObject.objects.create(name="test Obj")
        oid = orm_obj.jid
        newname = "TESTING new Name"
        orm_obj.name = newname
        orm_obj.save()

        loaded = models.JaseciObject.objects.get(name=newname)
        self.assertEqual(loaded.jid, oid)
        self.assertEqual(loaded.name, newname)

        orm_obj.delete()
        self.assertFalse(models.JaseciObject.objects.filter(name=newname).exists())

    def test_lookup_global_config(self):
        """Test look up config returns right value"""
        models.GlobalVars.objects.create(name="testname", value="testval")
        self.assertEqual(models.lookup_global_config("testname"), "testval")
        self.assertEqual(models.lookup_global_config("nonsense", "apple"), "apple")



import time

from django.db import connections
from django.db.utils import OperationalError
from django.core.management.base import BaseCommand
from jaseci.utils.utils import logger


class Command(BaseCommand):
    """Django command to wait for db up before starting Jaseci"""

    def handle(self, *args, **options):
        logger.info("Waiting for database...")
        db_conn = None
        while not db_conn:
            try:
                db_conn = connections["default"]
            except OperationalError:
                logger.info("Connection failed, retry in 1 seccond...")
                time.sleep(1)

        logger.info(self.style.SUCCESS("Database up!"))



from rest_framework import generics, permissions
from rest_framework.views import APIView
from rest_framework import renderers, status
from knox.views import LoginView as KnoxLoginView
from django.contrib.auth import login, get_user_model
from django.contrib.auth.signals import user_logged_out
from knox.auth import TokenAuthentication

from jaseci_serv.user_api.serializers import UserSerializer
from jaseci_serv.user_api.serializers import SuperUserSerializer
from jaseci_serv.user_api.serializers import AuthTokenSerializer
from jaseci_serv.user_api.serializers import send_activation_email
from jaseci_serv.base.models import lookup_global_config
from datetime import timedelta

from rest_framework.response import Response
import base64


class CreateUserView(generics.CreateAPIView):
    """Create a new user in the system"""

    serializer_class = UserSerializer

    def perform_create(self, serializer):
        """Perform create override to send out activation Email"""
        created_object = serializer.save()
        if not created_object.is_activated:
            send_activation_email(self.request, created_object.email)


class CreateSuperUserView(generics.CreateAPIView):
    """Create a new user in the system"""

    serializer_class = SuperUserSerializer


class ActivateUserView(APIView):
    """Create a new user in the system"""

    # TODO: Add Test case
    renderer_classes = [renderers.JSONRenderer]

    def get(self, request, code):
        """Perform create override to send out activation Email"""
        email = base64.b64decode(code).decode()
        user = get_user_model().objects.get(email=email)
        if not user:
            return Response("Invalid activation code/link!")
        user.is_activated = True
        user.save()
        return Response("User successfully activated!")


class CreateTokenView(KnoxLoginView):
    """Create a new auth token for user"""

    permission_classes = (permissions.AllowAny,)

    def get_token_ttl(self):
        ttl = lookup_global_config("LOGIN_TOKEN_TTL_HOURS", None)
        if not ttl:
            return super(CreateTokenView, self).get_token_ttl()
        else:
            return timedelta(hours=int(ttl))

    def post(self, request, format=None):
        serializer = AuthTokenSerializer(
            data=request.data, context={"request": request}
        )
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data["user"]
        login(request, user)
        return super(CreateTokenView, self).post(request, format=None)


class ManageUserView(generics.RetrieveUpdateAPIView):
    """Manage the authenticated user"""

    # TODO: re-activate when user changes email address or disable email change
    serializer_class = UserSerializer
    authentication_classes = (TokenAuthentication,)
    permission_classes = (permissions.IsAuthenticated,)

    def get_object(self):
        """Retrieve and return authenticated user"""
        return self.request.user

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)


class LogoutAllUsersView(APIView):
    """
    Log out of all user sessions across all users
    I.E. deletes all auth tokens for all users
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (permissions.IsAuthenticated, permissions.IsAdminUser)

    def post(self, request, format=None):
        users = get_user_model().objects.all()
        for u in users:
            u.auth_token_set.all().delete()
            user_logged_out.send(sender=u.__class__, request=request, user=u)
        return Response(None, status=status.HTTP_204_NO_CONTENT)


class IsSuperUser(permissions.BasePermission):
    """
    Allows access only to super users.
    """

    def has_permission(self, request, view):
        return bool(request.user and request.user.is_superuser)


class UpdateUserView(APIView):
    """
    update specific user's is_activated and is_superuser
    """

    serializer_class = UserSerializer
    authentication_classes = (TokenAuthentication,)
    permission_classes = (permissions.IsAuthenticated, IsSuperUser)

    def update_field(self, user, data: dict, field_map: dict):
        has_changes = False
        for field, _type in field_map.items():
            if field in data and type(data[field]) is _type:
                if getattr(user, field, _type()) != data[field]:
                    setattr(user, field, data[field])
                    has_changes = True

        return has_changes

    def post(self, request, id):
        user = get_user_model().objects.get(id=id)

        if user is None or not request.data or type(request.data) is not dict:
            return Response(status=status.HTTP_400_BAD_REQUEST)

        if self.update_field(
            user, request.data, {"is_activated": bool, "is_superuser": bool}
        ):
            user.save()
            return Response("Update Success!")
        else:
            return Response("No changes found!")



import base64

from django.contrib.auth import get_user_model
from django.urls import reverse
from jaseci_serv.utils.test_utils import skip_without_redis

from rest_framework.test import APIClient
from rest_framework import status

from jaseci.utils.utils import TestCaseHelper
from django.test import TestCase

import uuid
import os


class PublicJacApiTests(TestCaseHelper, TestCase):
    """Test the publicly available node API"""

    def setUp(self):
        super().setUp()
        self.client = APIClient()
        get_user_model().objects.create_user(
            "throwawayJSCITfdfdEST_test@jaseci.com", "password"
        )
        self.user = get_user_model().objects.create_user(
            "JSCITfdfdEST_test@jaseci.com", "password"
        )
        self.auth_client = APIClient()
        self.auth_client.force_authenticate(self.user)
        self.suser = get_user_model().objects.create_superuser(
            "JSCITfdfdEST_test2@jaseci.com", "password"
        )
        self.sauth_client = APIClient()
        self.sauth_client.force_authenticate(self.suser)

    def tearDown(self):
        super().tearDown()

    def test_login_required_jac_api(self):
        """Test that login required for retrieving nodes"""
        res = self.client.get(reverse("jac_api:graph_list"))

        self.assertEqual(res.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_public_jac_apis_walker_summon_auth(self):
        """Test public API for summoning walker"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_spawn_create", "name": "pubinit"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        wjid = res.data["jid"]
        payload = {"op": "walker_get", "mode": "keys", "wlk": wjid}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        key = res.data["anyone"]
        payload = {"op": "alias_list"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        nd = res.data["active:graph"]
        payload = {"op": "walker_summon", "key": key, "wlk": wjid, "nd": nd}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data["report"]), 0)
        key = "aaaaaaa"
        payload = {"op": "walker_summon", "key": key, "wlk": wjid, "nd": nd}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertFalse(res.data["success"])

    def test_serverside_sentinel_global_public_access_summon(self):
        """Test master delete operation"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "sentinel_active_global"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertTrue(res.data["success"])
        payload = {"op": "graph_create"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_spawn_create", "name": "pubinit"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_get", "mode": "keys", "wlk": "spawned:walker:pubinit"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        key = res.data["anyone"]
        payload = {"op": "alias_list"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        walk = res.data["spawned:walker:pubinit"]
        nd = res.data["active:graph"]
        payload = {"op": "walker_summon", "key": key, "wlk": walk, "nd": nd}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data["report"]), 0)

    def test_serverside_sentinel_global_public_access_callback(self):
        """Test public API for walker callback"""
        zsb_file = open(os.path.dirname(__file__) + "/public.jac").read()
        payload = {"op": "sentinel_register", "name": "public", "code": zsb_file}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "sentinel_active_global"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertTrue(res.data["success"])
        payload = {"op": "graph_create"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_spawn_create", "name": "callback"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_get", "mode": "keys", "wlk": "spawned:walker:callback"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        key = res.data["anyone"]
        payload = {"op": "alias_list"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        walk = res.data["spawned:walker:callback"].split(":")[2]
        nd = res.data["active:graph"].split(":")[2]
        payload = {"op": "walker_callback", "key": key, "wlk": walk, "nd": nd}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}', args=[payload["nd"], payload["wlk"]])
            + "?keys="
            + key,
            payload,
            format="json",
        )
        self.assertEqual(res.data["updated"], False)
        self.assertEqual(res.status_code, 200)
        zsb_file = open(os.path.dirname(__file__) + "/public_updated.jac").read()
        payload = {"op": "sentinel_register", "name": "public", "code": zsb_file}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 1)
        payload = {"op": "walker_callback", "key": key, "wlk": walk, "nd": nd}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}', args=[payload["nd"], payload["wlk"]])
            + "?keys="
            + key,
            payload,
            format="json",
        )
        self.assertEqual(res.data["updated"], True)
        self.assertEqual(res.status_code, 201)

    @skip_without_redis
    def test_serverside_sentinel_global_public_access_callback_async(self):
        """Test public API for walker callback"""
        zsb_file = open(os.path.dirname(__file__) + "/public.jac").read()
        payload = {"op": "sentinel_register", "name": "public", "code": zsb_file}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "sentinel_active_global"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertTrue(res.data["success"])
        payload = {"op": "graph_create"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_spawn_create", "name": "callback"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_get", "mode": "keys", "wlk": "spawned:walker:callback"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        key = res.data["anyone"]
        payload = {"op": "alias_list"}
        res = self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        walk = res.data["spawned:walker:callback"].split(":")[2]
        nd = res.data["active:graph"].split(":")[2]
        payload = {
            "is_async": True,
            "op": "walker_callback",
            "key": key,
            "wlk": walk,
            "nd": nd,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}', args=[payload["nd"], payload["wlk"]])
            + "?keys="
            + key,
            payload,
            format="json",
        )

        self.assertFalse("updated" in res.data)
        self.assertTrue("task_id" in res.data)

        task_id = res.data["task_id"]

        res = self.auth_client.get(
            reverse("jac_api:walker_queue_check") + f"?task_id={task_id}"
        )

        self.assertEqual("SUCCESS", res.data["status"])

        self.assertTrue(res.data["result"]["success"])


class PrivateJacApiTests(TestCaseHelper, TestCase):
    """Test the authorized user node API"""

    def setUp(self):
        super().setUp()
        get_user_model().objects.create_user(
            "throwawayJSCITfdfdEST_test@jaseci.com", "password"
        )
        self.user = get_user_model().objects.create_user(
            "JSCITfdfdEST_test@jaseci.com", "password"
        )
        self.suser = get_user_model().objects.create_superuser(
            "JSCITfdfdEST_test2@jaseci.com", "password"
        )
        self.client = APIClient()
        self.client.force_authenticate(self.user)
        self.sclient = APIClient()
        self.sclient.force_authenticate(self.suser)
        self.master = self.user.get_master()

    def tearDown(self):
        super().tearDown()

    def test_jac_api_create_graph(self):
        """Test API for creating a graph"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertEqual(res.status_code, status.HTTP_200_OK)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        self.assertEqual(gph.name, "root")

    def test_jac_api_create_sentinel(self):
        """Test API for creating a sentinel"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Untitled Sentinel"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        self.assertEqual(res.status_code, status.HTTP_200_OK)
        self.assertEqual(sent.name, "Untitled Sentinel")

    def test_jac_api_list_graphs_sents(self):
        """Test API for creating a sentinel"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {"op": "graph_list"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        glist = res.data
        payload = {"op": "sentinel_register", "name": "SomeSent"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {"op": "sentinel_register", "name": "SomeSent2"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {"op": "sentinel_list"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        slist = res.data
        self.assertEqual(len(glist), 1)
        self.assertEqual(len(slist), 2)
        self.assertEqual(slist[1]["name"], "SomeSent2")
        self.assertEqual(glist[0]["name"], "root")

    def test_jac_api_delete_graph(self):
        """Test API for deleteing a graph"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertIn(gph.id.urn, gph._h.mem.keys())
        payload = {"op": "graph_delete", "gph": gph.id.urn}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertNotIn(gph.id.urn, gph._h.mem.keys())

    def test_jac_api_get_graph_dot(self):
        """Test API for getting graph in dot str"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "graph_get", "mode": "dot", "gph": gph.id.urn, "dot": True}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertTrue("graph root" in res.json())

    def test_jac_api_delete_sentinel(self):
        """Test API for deleting a sentinel"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        self.assertIn(gph.id.urn, gph._h.mem.keys())
        self.assertIn(sent.id.urn, gph._h.mem.keys())
        payload = {"op": "sentinel_delete", "snt": sent.id.urn}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertIn(gph.id.urn, gph._h.mem.keys())
        self.assertNotIn(sent.id.urn, gph._h.mem.keys())

    def test_jac_api_get_jac_code(self):
        """Test API for deleting a sentinel"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {
            "op": "sentinel_register",
            "name": "Something",
            "code": 'walker testwalker{ std.log("hello"); }',
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {"op": "sentinel_get", "mode": "code", "snt": sent.id.urn}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertIn("walker test", res.data)

    def test_jac_api_sentinel_set(self):
        """Test API for deleting a sentinel"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": "node awesome;",
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertIn('{"name": "start"', sent.code_ir)

    def test_jac_api_sentinel_set_encoded(self):
        """Test API for deleting a sentinel"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        enc_str = base64.b64encode(b"node awesome;").decode()
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": enc_str,
            "encoded": True,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertIn('{"name": "start"', sent.code_ir)

    def test_jac_api_compile(self):
        """Test API for compiling a sentinel"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": 'walker testwalker{ std.log("hello"); }',
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertTrue(sent.is_active)
        self.assertEqual(sent.arch_ids.obj_list()[3].name, "testwalker")

    def test_jac_api_load_application(self):
        """Test API for loading an application"""
        enc_str = base64.b64encode(b'walker testwalker{ std.log("hello"); }').decode()
        payload = {
            "op": "sentinel_register",
            "name": "test_app",
            "code": enc_str,
            "encoded": True,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        snt = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        gph = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[1]["jid"])
        )
        self.assertEqual(snt.name, "test_app")
        self.assertEqual(gph.name, "root")
        self.assertTrue(snt.is_active)

    def test_jac_api_load_application_no_leak(self):
        """Test API for loading an application"""
        enc_str = base64.b64encode(
            b"node sample { has apple;} "
            + b"walker testwalker{ new = spawn here --> node::sample; "
            + b"report new; }"
        ).decode()
        payload = {
            "op": "sentinel_register",
            "name": "test_app",
            "code": enc_str,
            "encoded": True,
        }
        num_objs_a = len(self.master._h.mem.keys())
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")
        num_objs_b = len(self.master._h.mem.keys())
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")
        num_objs_c = len(self.master._h.mem.keys())
        self.assertLess(num_objs_a, num_objs_b)
        self.assertEqual(num_objs_b, num_objs_c)

    def test_jac_api_spawn(self):
        """Test API for spawning a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": 'walker testwalker{ std.log("hello"); }',
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "walker_spawn_create",
            "snt": sent.id.urn,
            "name": "testwalker",
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        walk = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        self.assertNotEqual(walk.id, sent.arch_ids.obj_list()[3].id)
        self.assertEqual(walk.name, sent.arch_ids.obj_list()[3].name)

    def test_jac_api_prime(self):
        """Test API for priming a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": 'walker testwalker{ std.log("hello"); }',
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "walker_spawn_create",
            "snt": sent.id.urn,
            "name": "testwalker",
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        walk = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {
            "op": "walker_prime",
            "wlk": walk.id.urn,
            "nd": gph.id.urn,
            "ctx": {},
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        self.assertIn(gph, walk.next_node_ids.obj_list())

    def test_jac_api_run(self):
        """Test API for running a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": 'walker testwalker{ std.log("hello"); }',
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "walker_spawn_create",
            "snt": sent.id.urn,
            "name": "testwalker",
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        walk = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {
            "op": "walker_prime",
            "wlk": walk.id.urn,
            "nd": gph.id.urn,
            "ctx": {},
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_execute", "wlk": walk.id.urn}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertGreater(walk.current_step, 0)

    def test_jac_api_graph_node_set(self):
        """Test API for setting context variables of node"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": "node sample { has apple;} "
            + "walker testwalker{ new = spawn here --> node::sample; "
            + "report new; }",
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "walker_spawn_create",
            "snt": sent.id.urn,
            "name": "testwalker",
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        walk = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {
            "op": "walker_prime",
            "wlk": walk.id.urn,
            "nd": gph.id.urn,
            "ctx": {},
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_execute", "wlk": walk.id.urn}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        nid = res.data["report"][0]["jid"]
        payload = {
            "op": "graph_node_set",
            "snt": sent.id.urn,
            "nd": nid,
            "ctx": {"apple": "TEST"},
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertGreater(walk.current_step, 0)
        self.assertEqual(res.data["context"]["apple"], "TEST")

    def test_deep_serialize_report(self):
        """Test API for running a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": "walker testwalker{ report " "[[[[[here, here], here], here]]]; }",
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "walker_spawn_create",
            "snt": sent.id.urn,
            "name": "testwalker",
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        walk = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {
            "op": "walker_prime",
            "wlk": walk.id.urn,
            "nd": gph.id.urn,
            "ctx": {},
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_execute", "wlk": walk.id.urn}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertGreater(walk.current_step, 0)

    def test_create_new_fields_auto_on_old_data(self):
        """Test API for running a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        gph = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {"op": "sentinel_register", "name": "Something"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        sent = self.master._h.get_obj(
            self.master.j_master, uuid.UUID(res.data[0]["jid"])
        )
        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": "node a { has b; } walker testwalker"
            + "{ r = spawn here --> node::a; r.b = 6; }",
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "walker_spawn_create",
            "snt": sent.id.urn,
            "name": "testwalker",
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        walk = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {
            "op": "walker_prime",
            "wlk": walk.id.urn,
            "nd": gph.id.urn,
            "ctx": {},
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_execute", "wlk": walk.id.urn}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertGreater(walk.current_step, 0)

        payload = {
            "op": "sentinel_set",
            "snt": sent.id.urn,
            "code": "node a { has b, c; } walker testwalker"
            + "{ with entry {take -->;} "
            + "a { here.c=7; std.log(here.c); }}",
            "encoded": False,
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "walker_spawn_create",
            "snt": sent.id.urn,
            "name": "testwalker",
        }
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        walk = self.master._h.get_obj(self.master.j_master, uuid.UUID(res.data["jid"]))
        payload = {
            "op": "walker_prime",
            "wlk": walk.id.urn,
            "nd": gph.id.urn,
            "ctx": {},
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_execute", "wlk": walk.id.urn}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.assertEqual(gph.outbound_nodes()[0].context["c"], 7)

    def test_master_create_linked_to_django_users(self):
        """Test master create operation"""
        payload = {
            "op": "master_create",
            "name": "yo@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertIn("j_type", res.data["user"])
        self.assertEqual(res.data["user"]["j_type"], "master")

    def test_master_create_linked_error_out(self):
        """Test master create operation"""
        payload = {
            "op": "master_create",
            "name": "yo@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "master_create",
            "name": "yo@gmail.com",
            "other_fields": {"password": "yoyo", "name": "", "is_activated": True},
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertFalse(res.data["success"])
        self.assertIn("already exists", res.data["response"])

    def test_master_create_linked_cant_override(self):
        """Test master create operation"""
        payload = {
            "op": "master_create",
            "name": "yo2@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyo",
                "name": "",
                "is_activated": True,
                "is_admin": True,
                "is_staff": True,
                "is_superuser": True,
            },
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(res.data["user"]["j_type"], "master")

    def test_master_create_linked_super_limited(self):
        """Test master create operation"""
        payload = {
            "op": "master_createsuper",
            "name": "yo3@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyo",
                "name": "",
                "is_activated": True,
                "is_admin": True,
                "is_staff": True,
                "is_superuser": True,
            },
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(res.status_code, 403)

    def test_master_create_linked_super_master_create(self):
        """Test master create operation"""
        payload = {
            "op": "master_createsuper",
            "name": "yo3@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyo",
                "name": "",
                "is_activated": True,
                "is_admin": True,
                "is_staff": True,
                "is_superuser": True,
            },
        }
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertIn("j_type", res.data["user"])
        self.assertEqual(res.data["user"]["j_type"], "super_master")

    def test_master_create_linked_to_django_users_login(self):
        """Test master create operation"""
        payload = {
            "op": "master_create",
            "name": "yo@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        login_client = APIClient()
        payload = {"email": "yo@gmail.com", "password": "yoyoyoyoyoyo"}
        res = login_client.post(reverse("user_api:token"), payload)
        self.assertIn("token", res.data)

    def test_master_create_linked_survives_orm(self):
        """Test master create operation"""
        self.user.get_master()._h.clear_cache()
        payload = {
            "op": "master_create",
            "name": "yo@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        login_client = APIClient()
        payload = {"email": "yo@gmail.com", "password": "yoyoyoyoyoyo"}
        res = login_client.post(reverse("user_api:token"), payload)
        self.assertIn("token", res.data)

    def test_master_create_super_inherets_from_django(self):
        """Test master create operation"""
        payload = {
            "op": "master_create",
            "name": "yo@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        login_client = APIClient()
        payload = {"email": "yo@gmail.com", "password": "yoyoyoyoyoyo"}
        res = login_client.post(reverse("user_api:token"), payload)
        self.assertIn("token", res.data)

    def test_master_delete_linked_to_django(self):
        """Test master delete operation"""
        payload = {
            "op": "master_create",
            "name": "yo@gmail.com",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        login_client = APIClient()
        payload = {"email": "yo@gmail.com", "password": "yoyoyoyoyoyo"}
        res = login_client.post(reverse("user_api:token"), payload)
        self.assertIn("token", res.data)
        payload = {"op": "master_delete", "name": "yo@gmail.com"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        login_client = APIClient()
        payload = {"email": "yo@gmail.com", "password": "yoyoyoyoyoyo"}
        res = login_client.post(reverse("user_api:token"), payload)
        self.assertNotIn("token", res.data)

    def test_serverside_sentinel_register_global(self):
        """Test master delete operation"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "sentinel_active_global"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertTrue(res.data["success"])
        payload = {"op": "sentinel_active_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertIn("jid", res.data)
        payload = {"op": "graph_list"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 0)
        payload = {"op": "graph_create"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 1)
        payload = {"op": "walker_run", "name": "init"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 3)

    def test_sentinel_active_global_with_auto_run(self):
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_create", "set_active": True}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "sentinel_active_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertFalse(res.data["success"])
        payload = {"op": "sentinel_active_global", "auto_run": "init"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertTrue(res.data["success"])
        self.assertIn("auto_run_result", res.data)
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 3)

    def test_public_user_create_global_init(self):
        public_client = APIClient()
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "user_create",
            "name": "yo@gmail.com",
            "global_init": "init",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = public_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(res.data["success"], True)
        self.assertEqual(res.data["global_init"]["auto_run_result"]["success"], True)

    def test_public_user_create_save_through(self):
        public_client = APIClient()
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "user_create",
            "name": "yo2@gmail.com",
            "global_init": "init",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = public_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        usr = get_user_model().objects.get(email="yo2@gmail.com")
        self.assertIsNotNone(usr.get_master().active_gph_id)

    def test_master_create_global_init(self):
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {
            "op": "master_create",
            "name": "yo@gmail.com",
            "global_init": "init",
            "other_fields": {
                "password": "yoyoyoyoyoyo",
                "name": "",
                "is_activated": True,
            },
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(res.data["success"], True)
        self.assertEqual(res.data["global_init"]["auto_run_result"]["success"], True)
        payload = {"op": "master_active_set", "name": "yo@gmail.com"}
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 3)

    def test_serverside_sentinel_unregister_global(self):
        """Test master delete operation"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "sentinel_active_global"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertTrue(res.data["success"])
        payload = {"op": "graph_create"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "init"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 3)
        payload = {"op": "graph_create"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 1)
        payload = {"op": "global_sentinel_unset"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.user._h.clear_cache()
        payload = {"op": "walker_run", "name": "init"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 1)

    def test_serverside_sentinel_register_pull(self):
        """Test master delete operation"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 2)
        payload = {"op": "global_sentinel_set"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "sentinel_pull"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertTrue(res.data["success"])
        payload = {"op": "sentinel_active_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertIn("jid", res.data)
        payload = {"op": "graph_list"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 0)
        payload = {"op": "graph_create"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 1)
        payload = {"op": "walker_run", "name": "init"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 3)

    def test_check_json_global_dict(self):
        """Test set get global objects (as json)"""
        from jaseci.tests.jac_test_code import set_get_global_dict

        payload = {
            "op": "sentinel_register",
            "name": "zsb",
            "code": set_get_global_dict,
        }
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "setter"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "getter"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(res.data["report"][0]["max_bot_count"], 10)

    def test_check_sentinel_set(self):
        """Test that sentinel set works serverside"""
        import jaseci.tests.jac_test_code as jtc

        payload = {
            "op": "sentinel_register",
            "name": "zsb",
            "auto_run": "",
            "code": jtc.version_label,
        }
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "sentinel_set", "code": jtc.set_get_global_dict}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "setter"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "getter"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(res.data["report"][0]["max_bot_count"], 10)

    def test_jac_walker_level_api_run(self):
        """Test API for running a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {
            "op": "sentinel_register",
            "name": "Something",
            "code": 'walker testwalker{ std.log("hello"); }',
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {}
        res = self.client.post(reverse("jac_api:walker_list"), payload)
        self.assertEqual(len(res.data), 1)
        res = self.client.post(reverse("jac_api:wapi", args=["testwalker"]), payload)
        self.assertTrue(res.data["success"])

    def test_jac_report_status_pluck(self):
        """Test API for running a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {
            "op": "sentinel_register",
            "name": "Something",
            "code": "walker ss {report:status = 302; report 4;}",
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {}
        res = self.client.post(reverse("jac_api:walker_list"), payload)
        self.assertEqual(len(res.data), 1)
        res = self.client.post(reverse("jac_api:wapi", args=["ss"]), payload)
        self.assertEqual(res.status_code, 302)
        self.assertTrue(res.data["success"])

    def test_jac_admin_master_allusers(self):
        """Test API for creating a graph"""
        payload = {}
        res = self.sclient.post(
            reverse("jac_api:master_allusers"), payload, format="json"
        )
        self.assertEqual(res.status_code, status.HTTP_200_OK)

    def test_asim_bug_check(self):
        """Test public API for summoning walker"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "asim_bug_check1"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        ret1 = res.data
        self.user.get_master()._h.clear_cache()
        payload = {"op": "walker_run", "name": "asim_bug_check2"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        ret2 = res.data
        self.assertEqual(ret1, ret2)

    def test_asim_bug_check2(self):
        """Test public API for summoning walker"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "asim_bug_check3"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        ret1 = res.data
        self.user.get_master()._h.clear_cache()
        payload = {"op": "walker_run", "name": "asim_bug_check2"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        ret2 = res.data
        self.assertEqual(ret1, ret2)

    def test_asim_bug_check3(self):
        """Test public API for summoning walker"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "asim_bug_check4"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        ret1 = res.data
        self.user.get_master()._h.clear_cache()
        payload = {"op": "walker_run", "name": "asim_bug_check2"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        ret2 = res.data
        self.assertEqual(ret1, ret2)

    def test_serverside_superuser_become(self):
        payload = {"op": "master_active_get"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        user_id = res.data["jid"]
        payload = {"op": "graph_create"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        user_gph_id = res.data["jid"]
        payload = {"op": "master_active_get"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        super_id = res.data["jid"]
        self.assertNotEqual(user_id, super_id)
        payload = {"op": "graph_list"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(len(res.data), 0)
        payload = {"op": "master_become", "mast": user_id}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "graph_list"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(user_gph_id, res.data[0]["jid"])
        payload = {"op": "master_active_get"}
        res = self.sclient.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        self.assertEqual(user_id, res.data["jid"])

    def test_jac_report_custom(self):
        """Test API for running a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {
            "op": "sentinel_register",
            "name": "Something",
            "code": "walker testwalker{ report:status = 302; "
            "report:custom = {'a': 'b'}; }",
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {}
        res = self.client.post(reverse("jac_api:walker_list"), payload)
        self.assertEqual(len(res.data), 1)
        res = self.client.post(reverse("jac_api:wapi", args=["testwalker"]), payload)
        self.assertEqual(res.data, {"a": "b"})
        self.assertEqual(res.status_code, status.HTTP_302_FOUND)

    def test_jac_create_user_in_jac_django(self):
        """Test API for running a walker"""
        payload = {"op": "graph_create"}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {
            "op": "sentinel_register",
            "name": "Something",
            "code": "walker testwalker{ report jaseci.master_create('a@b.com', '',  {},"
            "{'password': 'yoyoyoyoyoyo', 'name': '', 'is_activated': true}); }",
        }
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {}
        res = self.client.post(reverse("jac_api:walker_list"), payload)
        self.assertEqual(len(res.data), 1)
        res = self.client.post(reverse("jac_api:wapi", args=["testwalker"]), payload)
        self.assertIn("jid", res.data["report"][0]["user"].keys())
        self.assertEqual(res.data["report"][0]["user"]["name"], "a@b.com")

    def test_global_ref(self):
        """Test global action triggers"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        payload = {"op": "walker_run", "name": "global_actions"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data

        self.assertFalse(res["success"])
        self.assertEquals(payload, res["report"][4]["body"])

    def test_multipart_json_file(self):
        """Test multipart using json file as ctx parameter"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")
        with open(os.path.dirname(__file__) + "/test.json", "rb") as ctx:
            form = {
                "name": "simple",
                "ctx": ctx,
                "nd": "active:graph",
                "snt": "active:sentinel",
            }
            res = self.client.post(reverse(f'jac_api:{"walker_run"}'), data=form).data

        self.assertTrue(res["success"])
        self.assertEquals({"sample": "sample"}, res["report"][0]["ctx"])

    @skip_without_redis
    def test_multipart_json_file_async(self):
        """Test multipart using json file as ctx parameter"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")
        with open(os.path.dirname(__file__) + "/test.json", "rb") as ctx:
            form = {
                "name": "simple",
                "ctx": ctx,
                "nd": "active:graph",
                "snt": "active:sentinel",
            }
            res = self.client.post(
                reverse(f'jac_api:{"walker_run"}') + "?is_async=true", data=form
            ).data

        self.assertTrue("task_id" in res)
        task_id = res["task_id"]

        res = self.client.get(
            reverse("jac_api:walker_queue_wait") + f"?task_id={task_id}"
        ).data

        self.assertEqual("SUCCESS", res["status"])
        self.assertEquals({"sample": "sample"}, res["result"]["report"][0]["ctx"])

        res = self.client.get(
            reverse("jac_api:walker_queue_check") + f"?task_id={task_id}"
        ).data

        self.assertEqual("SUCCESS", res["status"])
        self.assertEquals({"sample": "sample"}, res["result"]["report"][0]["ctx"])

    def test_multipart_json_string(self):
        """Test multipart using json string as ctx parameter"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")

        form = {
            "name": "simple",
            "ctx": '{"sample":"sample"}',
            "nd": "active:graph",
            "snt": "active:sentinel",
        }

        res = self.client.post(reverse(f'jac_api:{"walker_run"}'), data=form).data

        self.assertTrue(res["success"])
        self.assertEquals({"sample": "sample"}, res["report"][0]["ctx"])

    def test_multipart_with_additional_file(self):
        """Test multipart with additional file"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")
        with open(os.path.dirname(__file__) + "/test.json", "rb") as ctx, open(
            os.path.dirname(__file__) + "/test.json", "rb"
        ) as ctx2:
            form = {
                "name": "simple_with_file",
                "ctx": ctx,
                "nd": "active:graph",
                "snt": "active:sentinel",
                "fileTypeField": ctx2,
            }
            res = self.client.post(reverse(f'jac_api:{"walker_run"}'), data=form).data

        default_file = [
            {
                "name": "test.json",
                "base64": "eyJzYW1wbGUiOiJzYW1wbGUifQ==",
                "content-type": "application/json",
            }
        ]

        self.assertTrue(res["success"])
        self.assertEquals(default_file, res["report"][0])
        self.assertEquals(default_file, res["report"][1])

    def test_multipart_custom_payload_with_additional_file(self):
        """Test multipart custom payload (non ctx format) with additional file"""
        zsb_file = open(os.path.dirname(__file__) + "/zsb.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")
        with open(os.path.dirname(__file__) + "/test.json", "rb") as ctx:
            form = {
                "name": "simple_custom_payload_with_file",
                "ctx": "",
                "nd": "active:graph",
                "snt": "active:sentinel",
                "fileTypeField": ctx,
            }
            res = self.client.post(reverse(f'jac_api:{"walker_run"}'), data=form).data

        default_file = [
            {
                "name": "test.json",
                "base64": "eyJzYW1wbGUiOiJzYW1wbGUifQ==",
                "content-type": "application/json",
            }
        ]

        self.assertTrue(res["success"])
        self.assertEquals(default_file, res["report"][2])

    def test_try_catch(self):
        """Test try catch triggers"""
        zsb_file = open(os.path.dirname(__file__) + "/try-syntax.jac").read()
        payload = {"op": "sentinel_register", "name": "zsb", "code": zsb_file}
        self.client.post(reverse(f'jac_api:{payload["op"]}'), payload, format="json")

        payload = {"op": "walker_run", "name": "walker_exception_no_try_else"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data

        self.assertIn(
            "in jac_try_exception\n    raise TryException(self.jac_exception(e, "
            "jac_ast))\njaseci.jac.machine.machine_state.TryException: ",
            res["stack_trace"],
        )

        self.assertIn(
            "zsb:walker_exception_no_try_else - line 6,"
            " col 20 - rule atom - Internal Exception: ",
            res["errors"][0],
        )

        payload = {"op": "walker_run", "name": "walker_exception_with_try_else"}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data

        self.assertFalse("stack_trace" in res)
        self.assertEqual("walker_exception_with_try_else", res["report"][0]["name"])
        self.assertEqual(14, res["report"][0]["line"])
        self.assertEqual(23, res["report"][0]["col"])
        self.assertIn(
            "zsb:walker_exception_with_try_else -"
            " line 14, col 23 - rule atom_trailer - ",
            res["errors"][0],
        )

        payload = {
            "op": "walker_run",
            "name": "walker_exception_with_try_else_multiple_line",
        }

        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data

        self.assertFalse("stack_trace" in res)
        self.assertEqual(
            "walker_exception_with_try_else_multiple_line", res["report"][0]["name"]
        )
        self.assertEqual(32, res["report"][0]["line"])
        self.assertEqual(23, res["report"][0]["col"])
        self.assertIn(
            "zsb:walker_exception_with_try_else_multiple_line "
            "- line 32, col 23 - rule atom_trailer - ",
            res["errors"][0],
        )

    def quick_call(self, bywho, ops):
        return bywho.post(reverse(f'jac_api:{ops["op"]}'), ops, format="json")

    def test_walker_smart_yield(self):
        testfile = open(os.path.dirname(__file__) + "/various.jac").read()
        self.quick_call(
            self.client, {"op": "sentinel_register", "name": "test", "code": testfile}
        )
        ret = self.quick_call(self.client, {"op": "walker_run", "name": "smart_yield"})
        ret = self.quick_call(self.client, {"op": "walker_run", "name": "smart_yield"})
        ret = self.quick_call(self.client, {"op": "walker_run", "name": "smart_yield"})
        ret = self.quick_call(self.client, {"op": "walker_run", "name": "smart_yield"})
        self.assertEqual(ret.data["report"], [{"id": 2}])

    def test_walker_smart_yield_no_future(self):
        testfile = open(os.path.dirname(__file__) + "/various.jac").read()
        self.quick_call(
            self.client, {"op": "sentinel_register", "name": "test", "code": testfile}
        )
        ret = self.quick_call(
            self.client, {"op": "walker_run", "name": "smart_yield_no_future"}
        )
        ret = self.quick_call(
            self.client, {"op": "walker_run", "name": "smart_yield_no_future"}
        )
        ret = self.quick_call(
            self.client, {"op": "walker_run", "name": "smart_yield_no_future"}
        )
        ret = self.quick_call(
            self.client, {"op": "walker_run", "name": "smart_yield_no_future"}
        )
        self.assertEqual(ret.data["report"], [{}])



import uuid
from datetime import datetime

from django.contrib.auth import get_user_model
from django.contrib.auth.models import (
    AbstractBaseUser,
    BaseUserManager,
    PermissionsMixin,
)
from django.db import models

from jaseci.api.interface import Interface
from jaseci.element.master import Master as CoreMaster
from jaseci.element.super_master import SuperMaster as CoreSuper
from jaseci_serv.jaseci_serv.settings import JASECI_CONFIGS
from jaseci_serv.svc import MetaService


class Master(CoreMaster):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._valid_configs += JASECI_CONFIGS

    def user_creator(self, name, other_fields: dict = {}):
        """
        Create a master instance and return root node master object

        other_fields used for additional fields for overloaded interfaces
        (i.e., Django interface)
        """
        data = {"email": name}
        for i in other_fields.keys():
            data[i] = other_fields[i]
        from jaseci_serv.user_api.serializers import UserSerializer

        serializer = UserSerializer(data=data)
        if serializer.is_valid(raise_exception=False):
            mas = serializer.save().get_master()
            mas._h = self._h
            return mas

    def superuser_creator(self, name, other_fields: dict = {}):
        """
        Create a master instance and return root node master object

        other_fields used for additional fields for overloaded interfaces
        (i.e., Django interface)
        """
        data = {"email": name}
        for i in other_fields.keys():
            data[i] = other_fields[i]
        from jaseci_serv.user_api.serializers import SuperUserSerializer

        serializer = SuperUserSerializer(data=data)
        if serializer.is_valid(raise_exception=False):
            mas = serializer.save().get_master()
            mas._h = self._h
            return mas

    def user_destroyer(self, name: str):
        """
        Permanently delete master with given id
        """
        get_user_model().objects.get(email=name).delete()


class SuperMaster(Master, CoreSuper):
    @Interface.admin_api()
    def master_allusers(self, limit: int = 10, offset: int = 0, asc: bool = False):
        """
        Returns info on a set of users, limit specifies the number of users to
        return and offset specfies where to start
        """

        if (limit < 0) or (offset < 0):
            return {"response": "Error occured! Parameters must be `positive numbers`!"}
        users = get_user_model().objects.all()

        if not asc:
            users = users.order_by("-time_created")
        total = users.count()
        end = offset + limit if limit else total
        filtered_users = []

        for i in users[offset:end]:
            filtered_users.append(
                {
                    "id": i.id,
                    "user": i.email,
                    "jid": i.master.urn,
                    "name": i.name,
                    "created_date": i.time_created.isoformat(),
                    "is_activated": i.is_activated,
                    "is_superuser": i.is_superuser,
                }
            )
        ret = {"total": total, "data": filtered_users}

        return ret


class UserManager(BaseUserManager):
    """
    Custom User Manager for Jaseci

    Note: Every user is linked to a single root node that is created upon the
    creation of the user.
    """

    def create_user(self, email, password=None, **extra_fields):
        """Creates and saves a new user"""
        # Makes first user admin
        if not get_user_model().objects.filter(is_admin=True).exists():
            return self.create_superuser(email, password, **extra_fields)
        email = self.normalize_email(email).lower()
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)

        # Create user's root node
        user.master = Master(h=user._h, name=email).id
        user._h.commit()

        user.save(using=self._db)
        return user

    def create_superuser(self, email, password=None, **extra_fields):
        """Creates and saves a new super user"""
        email = self.normalize_email(email).lower()
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.is_staff = True
        user.is_activated = True
        user.is_admin = True
        user.is_superuser = True
        user.save(using=self._db)

        # Create user's root node
        user.master = SuperMaster(h=user._h, name=email).id
        user._h.commit()

        user.save(using=self._db)
        return user


class User(AbstractBaseUser, PermissionsMixin):
    """
    Custom User Model to use email instead of username

    Root node  is attached to each User and created at user creation
    """

    email = models.EmailField(max_length=255, unique=True)
    name = models.CharField(max_length=255, blank=True)
    time_created = models.DateTimeField(auto_now_add=True)
    time_modified = models.DateTimeField(auto_now=True)
    is_active = models.BooleanField(default=True)
    is_activated = models.BooleanField(default=False)
    is_staff = models.BooleanField(default=False)
    is_admin = models.BooleanField(default=False)
    is_superuser = models.BooleanField(default=False)
    master = models.UUIDField(default=uuid.uuid4)
    objects = UserManager()

    def __init__(self, *args, **kwargs):
        self._h = MetaService().hook()
        AbstractBaseUser.__init__(self, *args, **kwargs)
        PermissionsMixin.__init__(self, *args, **kwargs)

    USERNAME_FIELD = "email"

    def get_master(self):
        """Returns main user Jaseci node"""
        return self._h.get_obj(caller_id=self.master.urn, item_id=self.master)

    def delete(self):
        JaseciObject.objects.filter(j_master=self.master.urn).delete()
        JaseciObject.objects.filter(jid=self.master.urn).delete()
        super().delete()


class JaseciObject(models.Model):
    """
    Generalized object model for Jaseci object types

    There is one table in db for all object types in the Jaseci machine
    which include nodes, edges, actions, contexts, walkers, and
    sentinels.

    Keep in mind Django's ORM can support many to many relationships
    between recursive schema's which may be useful if necessary in the
    future.
    """

    jid = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    j_parent = models.UUIDField(null=True, blank=True)
    j_master = models.UUIDField(null=True, blank=True)
    name = models.CharField(max_length=255, blank=True)
    kind = models.CharField(max_length=255, blank=True)
    j_timestamp = models.DateTimeField(default=datetime.utcnow)
    j_type = models.CharField(max_length=15, default="node")
    j_access = models.CharField(max_length=15, default="private")
    j_r_acc_ids = models.TextField(blank=True)
    j_rw_acc_ids = models.TextField(blank=True)
    jsci_obj = models.TextField(blank=True)


class GlobalVars(models.Model):
    """Global configuration item"""

    name = models.CharField(max_length=31, unique=True)
    value = models.TextField(blank=True)


def lookup_global_config(name, default=None):
    """Helper for looking up GlobalVars, returns default if not found"""
    try:
        return GlobalVars.objects.get(name=name).value
    except GlobalVars.DoesNotExist:
        return default



from jaseci.utils.utils import TestCaseHelper
from django.test import TestCase
from django.contrib.auth import get_user_model
from django.urls import reverse
from django.test import Client


class AdminTests(TestCaseHelper, TestCase):
    def setUp(self):
        """Creates test client, admin user and user for this test case"""
        super().setUp()
        self.client = Client()
        self.admin_user = get_user_model().objects.create_superuser(
            email="JSCITEST_admin@jaseci.com", password="password"
        )
        self.client.force_login(self.admin_user)
        self.user = get_user_model().objects.create_user(
            email="JSCITEST_test@jaseci.com", password="password"
        )

    def tearDown(self):
        """Deletes test users out of databases"""
        self.admin_user.delete()
        self.user.delete()
        super().tearDown()

    def test_users_listed(self):
        """Tests that users show up on admin page"""
        url = reverse("admin:base_user_changelist")
        res = self.client.get(url)

        self.assertTrue(self.user.name in str(res.content))
        self.assertTrue(self.user.email in str(res.content))
        self.assertEqual(res.status_code, 200)

    def test_user_add_edit_page(self):
        """Test that the user add and edit pages work"""
        url = reverse("admin:base_user_change", args=[self.user.id])
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)

        url = reverse("admin:base_user_add")
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)

    def test_base_has_global_config(self):
        """Tests that global config show up on admin page"""
        url = reverse("admin:base_globalvars_changelist")
        res = self.client.get(url)

        self.assertTrue("Global vars" in str(res.content))
        self.assertEqual(res.status_code, 200)

    def test_base_has_jaseci_objs(self):
        """Tests that jaseci objects show up on admin page"""
        url = reverse("admin:base_jaseciobject_changelist")
        res = self.client.get(url)

        self.assertTrue("Jaseci objects" in str(res.content))
        self.assertEqual(res.status_code, 200)



from django.contrib.auth import get_user_model
from django.db.models import Q

from jaseci.utils.utils import TestCaseHelper
from django.test import TestCase

from jaseci_serv.base.models import GlobalVars

# Alias for create user
create_user = get_user_model().objects.create_user
get_user = get_user_model().objects.get


class OrmConfigTestsPrivate(TestCaseHelper, TestCase):
    """Test Jaseci Engine when authenticated"""

    def setUp(self):
        super().setUp()
        self.user = create_user(
            email="JSCadft@jaseci.com",
            password="testpass",
            name="some dude",
        )

    def tearDown(self):
        super().tearDown()

    def test_jsci_db_to_engine_hook_saving(self):
        """Test that db hooks are set up correctly for saving"""
        user = self.user
        self.assertIsNotNone(user._h)
        h = user._h
        h.save_glob("GOOBY", "MOOBY")
        h.commit()

        load_test = GlobalVars.objects.get(name="GOOBY")
        self.assertEqual(load_test.value, "MOOBY")

    def test_jsci_db_to_engine_hook_loading(self):
        """Test that db glob hooks are set up correctly for loading"""
        user = self.user
        self.assertIsNotNone(user._h)
        h = user._h
        h.save_glob("GOOBY1", "MOOBY1")
        h.commit()

        del user._h.mem["global"]["GOOBY1"]
        if h.redis.is_running():
            h.redis.delete("GOOBY1")
        self.assertNotIn("GOOBY1", user._h.mem["global"].keys())

        load_test = GlobalVars.objects.filter(name="GOOBY1").first()
        self.assertIsInstance(load_test, GlobalVars)

        user._h.get_glob("GOOBY1")
        self.assertEqual(user._h.mem["global"]["GOOBY1"], "MOOBY1")

    def test_jsci_db_glob_listing(self):
        """Test that db glob hooks are set up correctly for listing"""
        user = self.user
        self.assertIsNotNone(user._h)
        h = user._h
        h.save_glob("GOOBY1", "MOOBY1")
        h.save_glob("GOOBY2", "MOOBY2")
        h.save_glob("GOOBY3", "MOOBY3")
        h.commit()

        h.clear_cache()
        if h.redis.is_running():
            h.redis.delete("GOOBY1")
        self.assertEqual(len(user._h.mem["global"].keys()), 0)
        self.assertNotIn("GOOBY1", user._h.mem["global"].keys())

        # Removing default configs
        GlobalVars.objects.filter(
            Q(name="REDIS_CONFIG") | Q(name="TASK_CONFIG") | Q(name="MAIL_CONFIG")
        ).delete()

        li = user._h.list_glob()
        self.assertEqual(len(li), 3)
        self.assertEqual(len(user._h.mem["global"].keys()), 0)
        user._h.get_glob("GOOBY1")
        self.assertEqual(user._h.mem["global"]["GOOBY1"], "MOOBY1")



"""
This module includes code related to hooking Jaseci's django models to the
core engine.

FIX: Serious permissions work needed
"""
import uuid

from django.core.exceptions import ObjectDoesNotExist

import jaseci as core_mod
from jaseci.hook import RedisHook
from jaseci.utils import utils
from jaseci.utils.json_handler import json_str_to_jsci_dict
from jaseci.utils.utils import logger


class OrmHook(RedisHook):
    """
    Hooks Django ORM database for Jaseci objects to Jaseci's core engine.
    """

    def __init__(self, objects, globs):
        self.objects = objects
        self.globs = globs
        super().__init__()

    ####################################################
    #                DATASOURCE METHOD                 #
    ####################################################

    # --------------------- OBJ ---------------------- #

    def get_obj_from_store(self, item_id):
        loaded_obj = super().get_obj_from_store(item_id)
        if loaded_obj is None:
            try:
                loaded_obj = self.objects.get(jid=item_id)
            except ObjectDoesNotExist:
                logger.error(
                    str(f"Object {item_id} does not exist in Django ORM!"),
                    exc_info=True,
                )
                return None
            class_for_type = self.find_class_and_import(loaded_obj.j_type, core_mod)
            kwargs = {"h": self, "m_id": loaded_obj.j_master.urn, "auto_save": False}
            ret_obj = class_for_type(**kwargs)
            utils.map_assignment_of_matching_fields(ret_obj, loaded_obj)
            assert uuid.UUID(ret_obj.jid) == loaded_obj.jid

            # Unwind jsci_payload for fields beyond element object
            obj_fields = json_str_to_jsci_dict(loaded_obj.jsci_obj, ret_obj)
            for i in obj_fields.keys():
                setattr(ret_obj, i, obj_fields[i])
            self.commit_obj_to_cache(ret_obj)
            return ret_obj
        return loaded_obj

    def has_obj_in_store(self, item_id):
        """
        Checks for object existance in store
        """
        return super().has_obj_in_store(item_id) or (
            self.objects.filter(jid=item_id).count()
        )

    def destroy_obj_from_store(self, item):
        super().destroy_obj_from_store(item)
        try:
            self.objects.get(jid=item.id).delete()
        except ObjectDoesNotExist:
            # NOTE: Should look at this at some point
            # logger.error("Object does not exists so delete aborted!")
            pass

    # --------------------- GLOB --------------------- #

    def get_glob_from_store(self, name):
        """
        Get global config from externally hooked general store by name
        """

        glob = super().get_glob_from_store(name)
        if glob is None:
            try:
                glob = self.globs.get(name=name).value
            except ObjectDoesNotExist:
                logger.error(
                    str(f"Global {name} does not exist in Django ORM!"), exc_info=True
                )
                return None

            super().commit_glob_to_cache(name, glob)
            return glob
        return glob

    def has_glob_in_store(self, name):
        """
        Checks for global config existance in store
        """
        return super().has_glob_in_store(name) or self.globs.filter(name=name).count()

    def list_glob_from_store(self):
        """Get list of global config to externally hooked general store"""
        globs = super().list_glob_from_store()

        if not globs:
            return [entry["name"] for entry in self.globs.values("name")]

        return globs

    def destroy_glob_from_store(self, name):
        """Destroy global config to externally hooked general store"""
        super().destroy_glob_from_store(name)
        try:
            self.globs.get(name=name).delete()
        except ObjectDoesNotExist:
            pass

    ####################################################
    #                    COMMITTER                     #
    ####################################################

    def commit_obj(self, item):
        item_from_db, created = self.objects.get_or_create(jid=item.id)
        utils.map_assignment_of_matching_fields(item_from_db, item)
        item_from_db.jsci_obj = item.jsci_payload()
        item_from_db.save()

    def commit_glob(self, name, value):
        self.commit_glob_to_cache(name, value)
        item_from_db, created = self.globs.get_or_create(name=name)
        item_from_db.value = value
        item_from_db.save()

    def commit(self, skip_cache=False):
        """Write through all saves to store"""
        for i in self.save_obj_list:
            if not skip_cache:
                self.commit_obj_to_cache(i)
            self.commit_obj(i)
        self.save_obj_list = set()

        for k, v in self.save_glob_dict.items():
            self.commit_glob(k, v)
        self.save_glob_dict = {}

    ###################################################
    #                  CLASS CONTROL                  #
    ###################################################

    def find_class_and_import(self, j_type, core_mod):
        if j_type == "master":
            from jaseci_serv.base.models import Master

            return Master
        elif j_type == "super_master":
            from jaseci_serv.base.models import SuperMaster

            return SuperMaster
        else:
            return utils.find_class_and_import(j_type, core_mod)



from jaseci.svc import MetaService as Ms
from jaseci_serv.jaseci_serv.settings import RUN_SVCS
from jaseci_serv.svc import MailService, RedisService, TaskService


class MetaService(Ms):
    def hook(self):
        h = self.app["hook"]()
        if RUN_SVCS:
            h.redis = RedisService(h)
            h.task = TaskService(h)
            h.mail = MailService(h)
        return h

    def build_hook(self):
        from jaseci_serv.base.models import GlobalVars, JaseciObject
        from jaseci_serv.hook.orm import OrmHook

        return OrmHook(objects=JaseciObject.objects, globs=GlobalVars.objects)

    def build_master(self, *args, **kwargs):
        from jaseci_serv.base.models import Master

        return Master(*args, **kwargs)

    def build_super_master(self, *args, **kwargs):
        from jaseci_serv.base.models import SuperMaster

        return SuperMaster(*args, **kwargs)



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.utils.translation import gettext as _

from jaseci_serv.base import models


class UserAdmin(BaseUserAdmin):
    """
    Customized user listing for admin page
    """

    ordering = ["time_created"]
    list_display = ["email", "name", "time_created"]
    fieldsets = (
        (None, {"fields": ("email", "password")}),
        (_("Personal Info"), {"fields": ("name",)}),
        (
            _("Permissions"),
            {
                "fields": (
                    "is_active",
                    "is_activated",
                    "is_staff",
                    "is_admin",
                    "is_superuser",
                )
            },
        ),
        (_("Important dates"), {"fields": ("last_login",)}),
    )
    add_fieldsets = (
        (None, {"classes": ("wide",), "fields": ("email", "password1", "password2")}),
    )


admin.site.register(models.User, UserAdmin)


class JaseciObjectAdmin(admin.ModelAdmin):
    ordering = ["j_timestamp"]
    list_display = ("jid", "name", "j_type", "j_parent", "j_master", "j_timestamp")
    search_fields = [
        "jid",
        "name",
        "j_type",
        "j_parent",
        "j_master",
        "j_timestamp",
    ]


admin.site.register(models.JaseciObject, JaseciObjectAdmin)


class GlobalVarsAdmin(admin.ModelAdmin):
    ordering = ["name"]
    list_display = ("name", "value")
    search_fields = ["name", "value"]


admin.site.register(models.GlobalVars, GlobalVarsAdmin)
